import inspect
import itertools
import re
import uuid
import warnings
from datetime import date, datetime, time, timedelta
from typing import Any, Callable, Dict, List, Optional, Union

import astropy.units as u
import matplotlib.colors as mcolors
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytz
from astroplan import Observer
from astropy.coordinates import SkyCoord
from astropy.time import Time, TimeDelta
from plotly.subplots import make_subplots
from timezonefinder import TimezoneFinder


class Night:
    def __init__(self, night_date: date, observations_within: str, observer: Observer):
        """
        Initialize a new instance of the Night class.

        Parameters
        ----------
        night_date : date
            The date of the night.
        observations_within : {"civil", "nautical", "astronomical"}
            Within which twilight observations will be done. Can be one of "civil", "nautical",
            or "astronomical".
        observer : astroplan.Observer
            An astroplan.Observer object that defines where the telescope is located in the world.
        """
        self.night_date = night_date
        self.observations_within = observations_within
        # Check if the observations_within is valid
        valid_options = ["civil", "nautical", "astronomical"]
        if self.observations_within not in valid_options:
            raise ValueError(f"observations_within must be one of {valid_options}")
        self.observer = observer

        # Calculate the solar midnight for the night based on the observer's location
        self.solar_midnight = self.calculate_solar_midnight()
        # Get the sunset and sunrise times for the night
        self.sunset = self.observer.sun_set_time(self.solar_midnight, which="previous")
        self.sunrise = self.observer.sun_rise_time(self.solar_midnight, which="next")

        # Get the times for the different twilights
        self.civil_evening = self.observer.twilight_evening_civil(
            self.solar_midnight, which="previous"
        )
        self.nautical_evening = self.observer.twilight_evening_nautical(
            self.solar_midnight, which="previous"
        )
        self.astronomical_evening = self.observer.twilight_evening_astronomical(
            self.solar_midnight, which="previous"
        )
        # And the same for the morning
        self.civil_morning = self.observer.twilight_morning_civil(
            self.solar_midnight, which="next"
        )
        self.nautical_morning = self.observer.twilight_morning_nautical(
            self.solar_midnight, which="next"
        )
        self.astronomical_morning = self.observer.twilight_morning_astronomical(
            self.solar_midnight, which="next"
        )

        # Time ranges for the different twilights
        self.time_range_solar = np.linspace(self.sunset, self.sunrise, 300)
        self.time_range_civil = np.linspace(self.civil_evening, self.civil_morning, 300)
        self.time_range_nautical = np.linspace(
            self.nautical_evening, self.nautical_morning, 300
        )
        self.time_range_astronomical = np.linspace(
            self.astronomical_evening, self.astronomical_morning, 300
        )
        # Extended time range for the night, 5 hours before and after sunset / sunrise
        self.time_range_extended = np.linspace(
            self.sunset - TimeDelta(5 / 24, format="jd"),
            self.sunrise + TimeDelta(5 / 24, format="jd"),
            300,
        )
        # Now only use the jd value of the twilights
        self.civil_evening = self.civil_evening.jd
        self.nautical_evening = self.nautical_evening.jd
        self.astronomical_evening = self.astronomical_evening.jd
        self.astronomical_morning = self.astronomical_morning.jd
        self.nautical_morning = self.nautical_morning.jd
        self.civil_morning = self.civil_morning.jd

        # Define the night time range based on the chosen twilight as the start and end times
        # of the observable night
        if self.observations_within == "civil":
            self.obs_within_limits = np.array([self.civil_evening, self.civil_morning])
            self.night_time_range = self.time_range_civil
        elif self.observations_within == "nautical":
            self.obs_within_limits = np.array(
                [self.nautical_evening, self.nautical_morning]
            )
            self.night_time_range = self.time_range_nautical
        elif self.observations_within == "astronomical":
            self.obs_within_limits = np.array(
                [self.astronomical_evening, self.astronomical_morning]
            )
            self.night_time_range = self.time_range_astronomical

        # Save the duration of the night
        self.night_duration = (
            self.night_time_range[-1] - self.night_time_range[0]
        ).to_datetime()

        # Calculate the alt-az frame of the moon for the night
        self.moon_altaz_frame = self.observer.moon_altaz(self.night_time_range)

    def calculate_solar_midnight(self):
        """
        Calculate the solar midnight for the night.

        Returns
        -------
        solar_midnight : astropy.time.Time
            The calculated solar midnight in UTC.
        """
        # Use timezonefinder to get the timezone
        timezone_str = TimezoneFinder().timezone_at(
            lat=self.observer.latitude.value, lng=self.observer.longitude.value
        )

        # Define the observer's local timezone
        local_timezone = pytz.timezone(timezone_str)

        # Create a localized datetime object
        local_datetime = datetime.combine(self.night_date, time(hour=18))
        local_datetime = local_timezone.localize(local_datetime)

        # Convert local datetime to UTC
        utc_datetime = local_datetime.astimezone(pytz.utc)

        # Convert the UTC datetime to an astropy Time object
        time_utc = Time(utc_datetime)

        # Calculate solar midnight in UTC based on the converted UTC time
        solar_midnight = self.observer.midnight(time_utc, which="next")
        return solar_midnight

    def calculate_culmination_window(self, obs_list):
        """
        Calculate the culmination window for the night based on the given list of observations.

        Parameters
        ----------
        obs_list : List[Observation]
            A list of Observation objects.
        """

        # Check if the objects in obs_list have the attribute culmination_time
        if not all(hasattr(obs, "culmination_time") for obs in obs_list):
            raise AttributeError(
                "sky_path() has to be run for all observations before calling this method."
            )

        # Get the culmination times of all observations
        culm_times = np.array([obs.culmination_time for obs in obs_list])
        # Check the observation is observable during the night
        is_observable = np.array(
            [np.any(obs.night_airmasses < 1.8) for obs in obs_list]
        )

        # Calculate the start and end times of the culmination window
        culm_window_start = np.min(culm_times[is_observable])
        culm_window_end = np.max(culm_times[is_observable])
        self.culmination_window = (culm_window_start, culm_window_end)
        # return culmination_window

    def __str__(self):
        lines = [
            f"Night(Date: {self.night_date},",
            f"      Sunset: {self.sunset},",
            f"      Sunrise: {self.sunrise},",
            f"      Civil evening: {self.civil_evening},",
            f"      Nautical evening: {self.nautical_evening},",
            f"      Astronomical evening: {self.astronomical_evening},",
            f"      Civil morning: {self.civil_morning},",
            f"      Nautical morning: {self.nautical_morning},",
            f"      Astronomical morning: {self.astronomical_morning},",
            f"      Observations within: '{self.observations_within}')",
        ]

        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, Night):
            return False
        else:
            return self.night_date == other.night_date


class Instrument:
    def __init__(self, name: str, instrument_type: str = "", plot_color: str = None):
        """
        Initialize a new instance of the Instrument class.

        Parameters
        ----------
        name : str
            The name of the instrument.
        instrument_type : str, optional
            The type of the instrument. Default is an empty string: "".
        plot_color : str, optional
            The color to use when plotting an observation of this instrument. Must be a valid hex
            code. By default the colors will be chosen from the 'Pastel1' color pallette from
            matplotlib.
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string")
        self.name = name
        if not isinstance(instrument_type, str):
            raise TypeError("instrument_type must be a string")
        self.instrument_type = instrument_type
        # Check that plot_color is a valid hex color code if its not None
        if plot_color is not None:
            if not bool(re.search(re.compile("^#([A-Fa-f0-9]{6})$"), plot_color)):
                raise ValueError(
                    "plot_color must be a valid hex color code, e.g. '#FF0000'"
                )
        self.plot_color = plot_color

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        lines = [
            "Instrument(",
            f"    Name = {self.name}",
        ]
        if self.instrument_type:
            lines.append(f"    Type = {self.instrument_type}")
        if self.plot_color:
            lines.append(f"    Plot color = {self.plot_color}")
        return "\n".join(lines)

    def __eq__(self, other):
        if not isinstance(other, Instrument):
            return False
        else:
            return self.name == other.name


class Program:
    def __init__(
        self,
        progID: str,
        instrument: Instrument,
        priority: int,
        time_share_allocated: float = 0.0,
        plot_color: str = None,
    ):
        """
        Initialize a new instance of the Program class.

        Parameters
        ----------
        progID : int or str
            The program ID code.
        instrument : str
            The name of the instrument that the program uses.
        time_share_allocated : float within [0, 1]
            The time share allocated to the program as a percentage of total time.
            Must be between 0 and 1.
        priority : int
            The priority of the program. Must be between 0 and 3, where 0 is the highest priority
            and 3 is the lowest.
        priority_base : int, optional
            The base value for mapping the priority. Defaults to 1.
        priority_offset : float, optional
            The offset value for mapping the priority. Defaults to 0.1.
        plot_color : str, optional
            The color to use when plotting an observation of this program. Must be a valid hex code.
            By default the colors will be chosen from the 'Set2' color pallette from matplotlib.
        """
        if not isinstance(progID, str):
            raise TypeError("progID must be a string")
        self.progID = progID
        if not isinstance(instrument, Instrument):
            raise TypeError("instrument must be of type Instrument")
        self.instrument = instrument
        # Check that priority value is valid
        if priority < 0 or priority > 3:
            raise ValueError("Priority must be between 0 and 3")
        if not isinstance(priority, int):
            raise TypeError("Priority must be an integer")
        self.priority = priority
        # Check that plot_color is a valid hex color code if its not None
        if plot_color is not None:
            if not (bool(re.search(re.compile("^#([A-Fa-f0-9]{6})$"), plot_color))):
                raise ValueError(
                    "plot_color must be a valid hex color code, e.g. '#FF0000'"
                )
        self.plot_color = plot_color
        # Check that time_share_allocated is valid
        if not (0 <= time_share_allocated <= 1):
            raise ValueError("Time share must be between 0 and 1")
        self.time_share_allocated = time_share_allocated
        # Initialize the current time usage to the allocated time share
        self.set_current_time_usage(self.time_share_allocated)

    def set_current_time_usage(self, current_time_usage: float):
        """
        Update the time share for the program.

        Parameters
        ----------
        current_time_usage : float
            The current time share used by the program in percent of the total
        """
        # Value check
        if not (0 <= current_time_usage <= 1):
            raise ValueError("current_time_usage must be between 0 and 1")
        # Update the current time share and calculate the difference
        self.time_share_current = current_time_usage
        self.time_share_pct_diff = self.time_share_current - self.time_share_allocated

    def __str__(self) -> str:
        lines = [
            "Program(",
            f"    ID = {self.progID}",
            f"    Instrument = {self.instrument}",
            f"    Time allocated = {self.time_share_allocated}",
            f"    Priority = {self.priority})",
        ]
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()


class Merit:
    def __init__(
        self,
        name: str,
        func: Callable,
        merit_type: str,
        parameters: Dict[str, Any] = {},
    ):
        """
        Initialize a new instance of the Merit class.

        Parameters
        ----------
        name : str
            The name of the merit.
        func : Callable
            The function that computes the merit.
        merit_type : {"fairness", "veto", "efficiency"}
            The type of the merit. Can be one of "fairness", "veto", or "efficiency".
        parameters : Dict[str, Any], optional
            Custom parameters for the merit function. Defaults to {}. The keys of the dictionary
            must match the names of the parameters of the function.
        """
        self.name = name
        self.func = func  # The function that computes this merit
        self.description = self.func.__doc__
        self.merit_type = merit_type  # "veto" or "efficiency"
        self.parameters = parameters  # Custom parameters for this merit

        # Check that the merit type is valid
        if self.merit_type not in ["fairness", "veto", "efficiency"]:
            raise ValueError(
                f"Invalid merit type ({self.merit_type}). "
                "Valid types are 'fairness', 'veto' and 'efficiency'."
            )

        # Consistency checks between the given func and parameters
        # It checks that the required parameters are all there, and that there are no extra
        # parameters that are not part of the function
        required_func_parameters = []
        optional_func_parameters = []
        for name, param in inspect.signature(self.func).parameters.items():
            if param.default == inspect.Parameter.empty:
                required_func_parameters.append(name)
            else:
                optional_func_parameters.append(name)
        # Check that the first parameter is "observation"
        if not (required_func_parameters[0] == "observation"):
            raise KeyError("The first parameter has to be 'observation'")
        # Check that the given parameters match the required parameters of the function
        if not set(required_func_parameters[1:]).issubset(set(self.parameters.keys())):
            raise ValueError(
                f"The given parameters ({set(self.parameters.keys())}) don't match the "
                "required parameters of the given function "
                f"({set(required_func_parameters[1:])})"
            )
        # Check that there are no extra parameters that are not part of the function
        if not set(self.parameters.keys()).issubset(
            set(required_func_parameters + optional_func_parameters)
        ):
            raise KeyError(
                "There are given parameters that are not part of the given function"
            )

    def evaluate(self, observation, **kwargs) -> float:
        """
        Evaluate the function with the given observation and additional arguments.

        Parameters
        ----------
        observation : Observation
            The input observation.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        float: The evaluation result.
        """
        # Combine custom parameters and runtime arguments, then call the function
        all_args = {**self.parameters, **kwargs}
        return self.func(observation, **all_args)

    def __str__(self):
        return f"Merit({self.name}, {self.merit_type}, {self.parameters})"

    def __repr__(self):
        return self.__str__()


class Target:
    def __init__(
        self,
        name: str,
        program: Program,
        coords: SkyCoord,
        priority: Union[int, None] = None,
        comment: str = "",
    ):
        """
        Initialize a new instance of the Target class.

        Parameters
        ----------
        name : str
            The name of the target.
        prog : Program
            The Program object that the target belongs to.
        coords : SkyCoord
            The coordinates of the target.
        priority : int
            The priority of the target. Must be between 0 and 3, where 0 is the highest priority
            and 3 is the lowest.
        comment : str, optional
            A comment about the target directed to the observer. Defaults to an empty string.
        """
        # Value checks
        if not isinstance(name, str):
            raise TypeError("name must be a string")
        self.name = name
        if not isinstance(program, Program):
            raise TypeError("program must be of type Program")
        self.program = program
        if not isinstance(coords, SkyCoord):
            raise TypeError("coords must be of type SkyCoord")
        self.coords = coords
        self.ra_deg = coords.ra.deg
        self.dec_deg = coords.dec.deg
        # Check that priority value is valid
        if priority is not None:
            if not isinstance(priority, int):
                raise TypeError(
                    f"Priority must be an integer, given: {priority} ({type(priority)})"
                )
        if isinstance(priority, int):
            if (priority < 0) or (priority > 3):
                raise ValueError(f"Priority must be between 0 and 3, given: {priority}")
        self.priority = priority
        if not isinstance(comment, str):
            raise TypeError(f"comment must be a string, given: '{comment}'")
        self.comment = comment

        self.fairness_merits: List[Merit] = []  # List of all fairness merits
        self.efficiency_merits: List[Merit] = []  # List of all efficiency merits
        self.veto_merits: List[Merit] = []  # List to store veto merits

    def add_merit(self, merit: Merit):
        """
        Adds a merit to the corresponding list based on its merit type.

        Parameters
        ----------
        merit : Merit
            The merit object to be added
        """
        if not isinstance(merit, Merit):
            raise TypeError("merit must be of type Merit")
        if merit.merit_type == "fairness":
            self.fairness_merits.append(merit)
        elif merit.merit_type == "veto":
            self.veto_merits.append(merit)
        elif merit.merit_type == "efficiency":
            self.efficiency_merits.append(merit)

    def add_merits(self, merits: List[Merit]):
        """
        Adds a list of Merit objects to the instance.

        Parameters
        ----------
        merits : List[Merit]
            A list of Merit objects to be added
        """
        if not isinstance(merits, list):
            raise TypeError("merits must be a list")
        if not all(isinstance(merit, Merit) for merit in merits):
            raise TypeError("the objects in merits must be of type Merit")
        for merit in merits:
            self.add_merit(merit)

    def __str__(self):
        lines = [
            f"Target(Name: {self.name},",
            f"       Program: {self.program.progID},",
            f"       Coordinates: ({self.coords.ra.deg:.3f}, {self.coords.dec.deg:.3f}),",
            f"       Priority: {self.priority},",
            f"       Fairness Merits: {self.fairness_merits},",
            f"       Veto Merits: {self.veto_merits},",
            f"       Efficiency Merits: {self.efficiency_merits})",
        ]

        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, Target):
            return False
        else:
            return self.name == other.name


class Observation:
    def __init__(
        self,
        target: Target,
        exposure_time: float,
    ):
        """
        Initialize a new instance of the Observation class.

        Parameters
        ----------
        target : Target
            The Target object representing the target being observed.
        exposure_time : float
            The exposure time, or duration, of the observation in seconds.
        """
        if not isinstance(target, Target):
            raise TypeError("target must be of type Target")
        self.target = target
        # Check if exposure time makes sense
        if exposure_time <= 0:
            raise ValueError("exposure_time must be greater than 0")
        # Check if exposure time is less than 1 second
        if exposure_time < 1:
            warnings.warn("exposure_time is less than 1 second")
        # Check if exposure time is greater than 9 hours
        if exposure_time > 32400:
            warnings.warn("exposure_time is greater than 9 hours")
        self.exposure_time = exposure_time / 86400  # Convert to days for internal use
        # Telescope altitude limits. These are only used to calculate the minimum and maximum
        # altitudes of the target during the night. The actual observational limits are set by the
        # Altitude merit defined by the user.
        self.tel_alt_lower_lim = 10.0
        self.tel_alt_upper_lim = 90.0
        self.score: float = 0.0  # Initialize score to zero
        self.veto_merits: List[float] = []  # List to store veto merits
        self.unique_id = uuid.uuid4()  # Unique ID for the observation instance

    def set_night(self, night: Night):
        """
        Set the night for the observation.

        Parameters
        ----------
        night : Night
            The Night object representing the night during which the observation takes place.
        """
        self.night = night

    def set_start_time(self, start_time: float):
        """
        Set the start time of the observation.

        Parameters
        ----------
        start_time : float
            The start time of the observation in JD (Julian Date).
        """
        # Check night has been assigned
        if not hasattr(self, "night"):
            raise AttributeError("Night must be assigned to the observation first")
        # Check if the start time is within the night time range
        # if not (
        #     self.night.night_time_range[0].jd
        #     <= start_time
        #     <= self.night.night_time_range[-1].jd
        # ):
        #     raise ValueError("Start time must be within the night time range")

        self.start_time = start_time
        self.end_time = self.start_time + self.exposure_time

    def skypath(self):
        """
        Calculate the skypath of the target during the night.
        To be run at the start of the scheduling process.
        """
        # Create the AltAz frame for the observation during the night
        self.night_altaz_frame = self.target.coords.transform_to(
            self.night.observer.altaz(time=self.night.night_time_range)
        )

        # Get the altitudes and airmasses of the target during the night
        self.night_altitudes = self.night_altaz_frame.alt.deg
        self.night_azimuths = self.night_altaz_frame.az.deg
        self.night_airmasses = self.night_altaz_frame.secz
        # Update the altitudes and airmasses for the observation timerange
        self.update_alt_airmass()
        # Get the minimum altitude of the target during the night
        self.min_altitude = max(self.night_altitudes.min(), self.tel_alt_lower_lim)
        # Get the maximum altitude of the target during the night
        self.max_altitude = min(self.night_altitudes.max(), self.tel_alt_upper_lim)

        # Convert extended time range to AltAz frame and get the time of maximum altitude
        self.culmination_time = self.night.time_range_extended[
            np.argmax(
                self.target.coords.transform_to(
                    self.night.observer.altaz(time=self.night.time_range_extended)
                ).alt.deg
            )
        ].jd

        # Get the rise and set times of the target
        start_time_astropy = self.night.night_time_range[0]
        if self.night_altitudes[0] > self.tel_alt_lower_lim:
            # If the target is already up by night start, the rise time is "previous"
            self.rise_time = self.night.observer.target_rise_time(
                start_time_astropy,
                self.target.coords,
                horizon=self.tel_alt_lower_lim * u.deg,
                which="previous",
                n_grid_points=10,
            ).jd
            # and set time is "next"
            self.set_time = self.night.observer.target_set_time(
                start_time_astropy,
                self.target.coords,
                horizon=self.tel_alt_lower_lim * u.deg,
                which="next",
                n_grid_points=10,
            ).jd
        else:
            # If the target is not up by night start, the rise time is "next"
            self.rise_time = self.night.observer.target_rise_time(
                start_time_astropy,
                self.target.coords,
                horizon=self.tel_alt_lower_lim * u.deg,
                which="next",
                n_grid_points=10,
            ).jd
            # and set time is also "next"
            self.set_time = self.night.observer.target_set_time(
                start_time_astropy,
                self.target.coords,
                horizon=self.tel_alt_lower_lim * u.deg,
                which="next",
                n_grid_points=10,
            ).jd

    def fairness(self) -> float:
        """
        Calculate the fairness score of the target.

        The fairness score is calculated by multiplying the priority of the target+program
        with the product of the evaluations of all fairness merits associated with the target.

        Returns
        -------
        float
            The fairness score of the target.
        """
        if len(self.target.fairness_merits) == 0:
            return 1.0
        else:
            return np.prod(
                [merit.evaluate(self) for merit in self.target.fairness_merits]
            )

    def update_alt_airmass(self):
        """
        Update the altitude and airmass values throughout the observation based on the start time
        and exposure time of the observation.
        """
        # Find indices
        night_range_jd = self.night.night_time_range.value
        start_idx = np.searchsorted(night_range_jd, self.start_time, side="left")
        end_idx = np.searchsorted(night_range_jd, self.end_time, side="right")
        self.obs_time_range = night_range_jd[start_idx:end_idx]
        self.obs_altitudes = self.night_altitudes[start_idx:end_idx]
        self.obs_azimuths = self.night_azimuths[start_idx:end_idx]
        self.obs_airmasses = self.night_airmasses[start_idx:end_idx].value

    def feasible(self, verbose: bool = False) -> float:
        """
        Determines the feasibility of the target based on the veto merits.

        Parameters
        ----------
        verbose : bool, optional
            If True, prints the name and value of each veto merit. Defaults to False

        Returns
        -------
        float
            The sensibility value, which is the product of all veto merit values.
        """
        veto_merit_values = []
        for merit in self.target.veto_merits:
            value = merit.evaluate(self)
            veto_merit_values.append(value)
            if verbose:
                print(f"{merit.name}: {value}")
            if value == 0.0:
                break
        self.sensibility_value = np.prod(veto_merit_values)
        return self.sensibility_value  # type: ignore

    def evaluate_score(self, verbose: bool = False) -> float:
        """
        Evaluates the score of the observation based on fairness, sensibility, and efficiency.

        Parameters
        ----------
        verbose : bool, optional
            If True, print the fairness, sensibility, efficiency, and rank score.

        Returns
        -------
        float
            The score of the observation.
        """
        # --- Fairness ---
        # Balances time allocation and priority
        fairness = self.fairness()

        # --- Sensibility ---
        # Veto merits that check for observatbility
        # This is calculated in the feasible method which is called every time the observation is
        # considered for the schedule
        sensibility = self.sensibility_value

        # --- Efficiency ---
        # Efficiency merits that check for scientific goal
        if len(self.target.efficiency_merits) == 0:
            efficiency = 1.0
        else:
            efficiency = np.mean(
                [merit.evaluate(self) for merit in self.target.efficiency_merits]
            )

        # --- Rank Score ---
        # Calculate total rank score by taking the product of fairness, sensibility and efficiency
        self.score = np.prod([fairness, sensibility, efficiency])  # type: ignore

        # Print results if verbose
        if verbose:
            print(f"Fairness: {fairness}")
            print(f"Sensibility: {sensibility}")
            print(f"Efficiency: {efficiency}")
            print(f"Rank score: {self.score}")

        return self.score

    def update_start_and_score(self, start_time):
        """
        Update the start time of the observation and recalculate the score.

        Parameters
        ----------
        start_time : float
            The new start time to set for the observation
        """
        self.set_start_time(start_time)
        self.update_alt_airmass()
        self.feasible()
        self.evaluate_score()

    def __str__(self):
        lines = [
            f"Observation(Target: {self.target.name},",
            f"            Start time: {self.start_time},",
            f"            Exposure time: {self.exposure_time},",
            f"            Score: {self.score})",
        ]

        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, Observation):
            return False
        else:
            return self.unique_id == other.unique_id


class Overheads:
    """
    A class for managing the calculation of the overheads between two consecutive observations.

    This class handles the calculation of various overheads involved in transitioning from one
    observation position to another.
    """

    def __init__(
        self,
        slew_rate_az: float,
        slew_rate_alt: float,
        cable_wrap_angle: Optional[float] = None,
    ):
        """
        Initialize a new instance of the Overheads class. The slew rates are given in degrees per
        second.

        Parameters
        ----------
        slew_rate_az : float
            The azimuth slew rate in degrees per second.
        slew_rate_alt : float
            The altitude slew rate in degrees per second.
        cable_wrap_angle : float, optional
            The azimuth angle where the cable wrap limit is at, in degrees. Default is None.
        """
        # Validate that slew rates are positive
        if slew_rate_az <= 0 or slew_rate_alt <= 0:
            raise ValueError("Slew rates must be positive")
        # Validate that cable wrap angle is between 0 and 360 degrees
        if cable_wrap_angle is not None:
            if not (0 <= cable_wrap_angle <= 360):
                raise ValueError("Cable wrap angle must be between 0 and 360 degrees")

        self.slew_rate_az = slew_rate_az
        self.slew_rate_alt = slew_rate_alt
        self.cable_wrap_angle = cable_wrap_angle
        self.overheads = []

    def _validate_function_params(self, func):
        """
        Validate the parameters of the given function.

        Parameters
        ----------
        func : Callable
            The function to validate.

        Returns
        -------
        bool
            True if the function has exactly two parameters named 'observation1' and 'observation2',
            False otherwise.
        """
        params = inspect.signature(func).parameters
        param_names = list(params.keys())
        expected_names = ["observation1", "observation2"]

        # Check if the function has exactly two parameters named 'observation1' and 'observation2'
        if param_names == expected_names:
            return True
        else:
            return False

    def _is_angle_between(
        self, start_angle: float, end_angle: float, check_angle: float
    ) -> bool:
        """
        Check if an angle is between two other angles.

        Parameters
        ----------
        start_angle : float
            The start angle in degrees.
        end_angle : float
            The end angle in degrees.
        check_angle : float
            The angle to check if it is between the start and end angles.

        Returns
        -------
        bool
            True if the check angle is between the start and end angles, False otherwise.
        """
        # Normalize angles to ensure they are between 0 and 360 degrees
        start_angle %= 360
        end_angle %= 360
        check_angle %= 360

        # Find the normalized positions of end and check angles relative to the start angle
        end_relative = (end_angle - start_angle) % 360
        check_relative = (check_angle - start_angle) % 360

        # Check if the check_angle is between start_angle and end_angle on the shortest path
        return (
            check_relative <= end_relative
            if end_relative <= 180
            else check_relative >= end_relative
            or check_relative <= (end_relative - 360)
        )

    def add_overhead(self, overhead_func, can_overlap_with_slew: bool = False):
        """
        Add an overhead function to the list of overheads.

        The function must have exactly two parameters named 'observation1' and 'observation2'.

        Parameters
        ----------
        overhead_func : Callable
            The overhead function to be added.
        """
        # Check that the given object is a function
        if not inspect.isfunction(overhead_func):
            raise TypeError("overhead_func must be a function")
        # Check that the function has the correct parameters
        if not self._validate_function_params(overhead_func):
            raise ValueError(
                "Function must have exactly two parameters named 'observation1' and 'observation2'"
            )

        # Add the function to the list of overheads
        self.overheads.append((overhead_func, can_overlap_with_slew))

    def calculate_slew_time(self, obs1: Observation, obs2: Observation):
        """
        Calculate the slew time between two observations, taking into account cable wrap.

        Parameters
        ----------
        obs1 : Observation
            The first observation.
        obs2 : Observation
            The second observation.
        cable_wrap_angle : float
            The azimuth angle where the cable wrap limit is at, in degrees.

        Returns
        -------
        slew_time : float
            The slew time in seconds.
        """
        # Calculate the difference in altitude and azimuth between the two observations
        time_idx = np.searchsorted(
            obs1.night.night_time_range.jd, obs1.end_time, side="left"
        )
        obs1_alt = obs1.night_altitudes[time_idx]
        obs1_az = obs1.night_azimuths[time_idx]
        obs2_alt = obs2.night_altitudes[time_idx]
        obs2_az = obs2.night_azimuths[time_idx]

        abs_sep_alt = np.abs(obs1_alt - obs2_alt)
        abs_sep_az = np.abs(obs1_az - obs2_az)
        # Take shortest path if azimuth separation is larger than 180 degrees
        if abs_sep_az > 180:
            abs_sep_az = 360 - abs_sep_az

        # Check if cable wrap is between the two azimuths when taking the shortest path
        if self.cable_wrap_angle is not None:
            if self._is_angle_between(obs1_az, obs2_az, self.cable_wrap_angle):
                # If cable wrap is between the two azimuths, azimuth slew has to go through long way
                abs_sep_az = 360 - abs_sep_az

        slew_time_az = abs_sep_az / self.slew_rate_az
        slew_time_alt = abs_sep_alt / self.slew_rate_alt
        # As Az and Alt slew happens simultaneously, total slew time will be the largest of the two
        slew_time = max([slew_time_az, slew_time_alt]) / 86400  # Convert to days
        return slew_time

    def calculate_transition(self, observation1, observation2):
        """
        Calculate the total overhead time between two observations.

        Parameters
        ----------
        observation1 : Observation
            The first observation.
        observation2 : Observation
            The second observation.
        """
        # Calculate slew time based on RA and DEC differences and slew rates
        slew_time = self.calculate_slew_time(observation1, observation2)

        total_overhead = slew_time
        extra_overhead_time = 0.0

        # Calculate other overheads
        for overhead_func, can_overlap_with_slew in self.overheads:
            overhead_time = (
                overhead_func(observation1, observation2) / 86400
            )  # Convert to days

            if can_overlap_with_slew:
                total_overhead = np.max([total_overhead, overhead_time])
            else:
                extra_overhead_time += overhead_time

        total_overhead += extra_overhead_time

        return total_overhead

    def __str__(self) -> str:
        return f"Overheads(\n\tslew_rate_az={self.slew_rate_az}, \n\tslew_rate_alt={self.slew_rate_alt}, \n\tcable_wrap_limit={self.cable_wrap_angle})"


class Plan:
    """
    A container class for Observation objects representing a plan for scheduling observations.
    """

    def __init__(self):
        self.observations: List[Observation] = []
        self.score = 0.0
        self.evaluation = 0.0

    def add_observation(self, observation: Observation):
        """
        Add an observation to the plan.

        Parameters
        ----------
        observation : Observation
            The Observation object to be added to the plan.
        """
        self.observations.append(observation)
        return self

    def calculate_overhead(self):
        """
        Calculates the overheads for the entire plan, as well as the total observation time.
        """
        # Calculate the overheads for the plan
        # Go through all observation and count the time between the end of one observation and the
        # start of the next one
        if len(self) == 0:
            self.observation_time = 0.0
            self.overhead_time = 0.0
            self.unused_time = 0.0
            self.overhead_ratio = 0.0
            self.observation_ratio = 0.0
            return
        else:
            first_obs = self.observations[0]
            observation_time = np.sum([obs.exposure_time for obs in self.observations])
            # Check that overhead and observation time add up to the total time
            total_time = self.observations[-1].end_time - first_obs.start_time
            overhead_time = total_time - observation_time
            available_obs_time = (
                first_obs.night.obs_within_limits[1]
                - first_obs.night.obs_within_limits[0]
            )
            unused_time = available_obs_time - observation_time - overhead_time

            self.observation_time = timedelta(days=observation_time)
            self.overhead_time = timedelta(days=overhead_time)
            self.unused_time = timedelta(days=unused_time)
            self.overhead_ratio = overhead_time / total_time
            self.observation_ratio = observation_time / total_time

    def calculate_avg_airmass(self):
        """
        Calculate the average airmass of the observations in the plan.
        """
        self.avg_airmass = np.mean(
            [
                np.max(obs.obs_airmasses)
                for obs in self.observations
                if len(obs.obs_airmasses) > 0
            ]
        )

    def evaluate_plan(self, w_score: float = 0.2, w_overhead: float = 0.8) -> float:
        """
        Calculates the evaluation of the plan. This is the mean of the individual scores of all
        observations times the observation ratio of the plan. This is to compensate between maximum
        score of the observations but total observation time.

        Parameters
        ----------
        w_score : float, optional
            The weight of the score in the evaluation. Defaults to 0.4.
        w_overhead : float, optional
            The weight of the overhead in the evaluation. Defaults to 0.6.

        Returns
        -------
        float
            The evaluation of the plan.
        """
        if w_score + w_overhead != 1:
            raise ValueError("The weights must sum to 1")
        # Evaluate the whole observation plan
        self.score = float(
            np.mean([obs.score for obs in self.observations]) if len(self) > 0 else 0
        )  # type: ignore
        # Calculate the overheads for the plan
        self.calculate_overhead()
        # Calculate the average airmass of the observations in the plan
        self.calculate_avg_airmass()
        # Calculate the evaluation of the plan
        self.evaluation = w_score * self.score + w_overhead * self.observation_ratio
        return self.evaluation

    def print_stats(self):
        """Prints some general information and statistics of the plan."""
        self.evaluate_plan()
        print(f"Length = {len(self)}")
        print(f"Score = {self.score:.6f}")
        print(f"Evaluation = {self.evaluation:.6f}")
        print(f"Observation time = {self.observation_time}")
        print(f"Overhead time = {self.overhead_time}")
        print(f"Observation ratio = {self.observation_ratio:.5f}")
        print(f"Overhead ratio = {self.overhead_ratio:.5f}")
        print(f"Avg airmass = {self.avg_airmass:.5f}")

    def plot(self, display: bool = True, path: str = None):
        """
        Plot the schedule for the night.

        Parameters
        ----------
        display : bool, optional
            Option to display the plot. Defaults to True.
        path : str, optional
            The path to the file where the plot will be saved. Defaults to None.
        """
        first_obs = self.observations[0]

        # Get sunset and sunrise times for this night
        night = first_obs.night
        sunset = Time(night.sunset, format="jd")
        sunrise = Time(night.sunrise, format="jd")

        # Get the times for the different twilights
        civil_evening = Time(night.civil_evening, format="jd").datetime
        nautical_evening = Time(night.nautical_evening, format="jd").datetime
        astronomical_evening = Time(night.astronomical_evening, format="jd").datetime
        civil_morning = Time(night.civil_morning, format="jd").datetime
        nautical_morning = Time(night.nautical_morning, format="jd").datetime
        astronomical_morning = Time(night.astronomical_morning, format="jd").datetime

        # Get which programs are part of this plan
        programs = list(
            set([obs.target.program for obs in self.observations])
        )  # Remove duplicates
        # Define unique colors for each program
        # if the programs have their plot_color attribute set, use that color
        # otherwise use the default color ('Set2' color pallette from matplotlib)
        default_colors = itertools.cycle(
            [mcolors.rgb2hex(color) for color in plt.get_cmap("Set2").colors]
        )
        prog_colors = {}
        for prog in programs:
            if prog.plot_color is None:
                prog_colors[prog.progID] = next(default_colors)
            else:
                prog_colors[prog.progID] = prog.plot_color

        fig, ax1 = plt.subplots(figsize=(13, 5))

        # Plot the altitude tracks of the targets
        for i, obs in enumerate(self.observations):
            # TODO clean up this part by using existing variables in the obs objects
            solar_altaz_frame = obs.target.coords.transform_to(
                night.observer.altaz(time=night.time_range_solar)
            )
            solar_night_altitudes = solar_altaz_frame.alt.deg

            # Plot altitude tracks of the target
            # Through the entire night
            ax1.plot_date(
                Time(obs.night.time_range_solar, format="jd").datetime,
                solar_night_altitudes,
                "-.",
                c="gray",
                alpha=0.6,
                lw=0.3,
            )
            # Only the observed period in highlighted color
            ax1.plot_date(
                Time(obs.obs_time_range, format="jd").datetime,
                obs.obs_altitudes,
                "-",
                c=prog_colors[obs.target.program.progID],
                lw=2,
                solid_capstyle="round",
            )

        # TODO Plot the tracks of the moon

        # Plot shaded areas between sunset and civil, nautical, and astronomical evening
        y_range = np.arange(0, 91)
        alpha_tw_fill = 0.2
        ax1.fill_betweenx(
            y_range, sunset.datetime, civil_evening, color="yellow", alpha=alpha_tw_fill
        )
        ax1.fill_betweenx(
            y_range,
            civil_evening,
            nautical_evening,
            color="orange",
            alpha=alpha_tw_fill,
        )
        ax1.fill_betweenx(
            y_range,
            nautical_evening,
            astronomical_evening,
            color="red",
            alpha=alpha_tw_fill,
        )
        # Same for the morning
        ax1.fill_betweenx(
            y_range,
            civil_morning,
            sunrise.datetime,
            color="yellow",
            alpha=alpha_tw_fill,
        )
        ax1.fill_betweenx(
            y_range,
            nautical_morning,
            civil_morning,
            color="orange",
            alpha=alpha_tw_fill,
        )
        ax1.fill_betweenx(
            y_range,
            astronomical_morning,
            nautical_morning,
            color="red",
            alpha=alpha_tw_fill,
        )
        # Add text that have the words "civil", "nautical", and "astronomical".
        # These boxes are placed vertically at the times of each of them (both evening and morning)
        text_kwargs = {
            "rotation": 90,
            "verticalalignment": "bottom",
            "color": "gray",
            "fontsize": 8,
        }

        ax1.text(
            sunset.datetime, 30.5, "Sunset", horizontalalignment="right", **text_kwargs
        )
        ax1.text(
            civil_evening, 30.5, "Civil", horizontalalignment="right", **text_kwargs
        )
        ax1.text(
            nautical_evening,
            30.5,
            "Nautical",
            horizontalalignment="right",
            **text_kwargs,
        )
        ax1.text(
            astronomical_evening,
            30.5,
            "Astronomical",
            horizontalalignment="right",
            **text_kwargs,
        )
        ax1.text(
            (civil_morning + timedelta(minutes=3)),
            30.5,
            "Civil",
            horizontalalignment="left",
            **text_kwargs,
        )
        ax1.text(
            (nautical_morning + timedelta(minutes=3)),
            30.5,
            "Nautical",
            horizontalalignment="left",
            **text_kwargs,
        )
        ax1.text(
            (astronomical_morning + timedelta(minutes=3)),
            30.5,
            "Astronomical",
            horizontalalignment="left",
            **text_kwargs,
        )
        ax1.text(
            (sunrise.datetime + timedelta(minutes=3)),
            30.5,
            "Sunrise",
            horizontalalignment="left",
            **text_kwargs,
        )

        # Use DateFormatter to format x-axis to only show time
        time_format = mdates.DateFormatter("%H:%M")
        ax1.xaxis.set_major_formatter(time_format)

        # Set the major ticks to an hourly interval
        hour_locator = mdates.HourLocator(interval=1)
        ax1.xaxis.set_major_locator(hour_locator)

        # Add a legend at the bottom of the plot to identiy the program colors
        legend_elements = [
            plt.Line2D(
                [0],
                [0],
                color=prog_colors[prog.progID],
                lw=2,
                label=f"{prog.progID}",
            )
            for prog in programs
        ]
        ax1.legend(handles=legend_elements, loc="lower center", ncol=len(programs))

        # In the title put the date of the schedule
        plt.title(f"Schedule for the night of {self.observations[0].night.night_date}")
        ax1.set_xlabel("Time [UTC]")
        ax1.set_ylabel("Altitude [deg]")
        ax1.set_ylim(30, 90)

        # Add a second axis to show the airmass
        # Set up airmass values and compute the corresponding altitudes for those airmass values
        desired_airmasses = np.arange(1.8, 0.9, -0.1)
        corresponding_altitudes = list(
            90.0 - np.degrees(np.arccos(1.0 / desired_airmasses[:-1]))
        )
        corresponding_altitudes.append(90.0)

        # Create the secondary y-axis for Airmass
        ax2 = ax1.twinx()
        ax2.set_ylim(ax1.get_ylim())
        # Set y-ticks at computed altitudes for desired airmasses
        ax2.set_yticks(corresponding_altitudes)
        ax2.set_yticklabels(
            np.round(desired_airmasses, 2)
        )  # Display the desired airmass values
        ax2.set_ylabel("Airmass")
        ax2.tick_params("y")
        if path is not None:
            plt.tight_layout()
            plt.savefig(path, dpi=300)
        if display:
            plt.show()
        else:
            plt.close()

    def plot_interactive(self, path: str = None):
        """
        Makes an interactive plot of the Plan using Plotly.

        Parameters
        ----------
        path : str, optional
            The path to the file where the plot will be saved. Defaults to None.
        """
        # Import Plotly here to avoid unnecessary dependencies
        import plotly.graph_objs as go

        first_obs = self.observations[0]

        # Get sunset and sunrise times for this night
        night = first_obs.night
        sunset = Time(night.sunset, format="jd")
        sunrise = Time(night.sunrise, format="jd")

        # Get the times for the different twilights
        civil_evening = Time(night.civil_evening, format="jd").datetime
        nautical_evening = Time(night.nautical_evening, format="jd").datetime
        astronomical_evening = Time(night.astronomical_evening, format="jd").datetime
        civil_morning = Time(night.civil_morning, format="jd").datetime
        nautical_morning = Time(night.nautical_morning, format="jd").datetime
        astronomical_morning = Time(night.astronomical_morning, format="jd").datetime

        # Get which programs are part of this plan
        programs = list(
            set([obs.target.program for obs in self.observations])
        )  # Remove duplicates
        # Define unique colors for each program
        # if the programs have their plot_color attribute set, use that color
        # otherwise use the default color ('Set2' color pallette from matplotlib)
        default_colors_iter = itertools.cycle(
            [mcolors.rgb2hex(color) for color in plt.get_cmap("Set2").colors]
        )

        # Create a dictionary that maps each program to a color
        prog_colors = {}
        for prog in programs:
            if prog.plot_color is None:
                prog_colors[prog.progID] = next(default_colors_iter)
            else:
                prog_colors[prog.progID] = prog.plot_color

        # Create subplots
        fig = make_subplots(specs=[[{"secondary_y": True}]])

        # Track which programs have been added to the legend
        added_to_legend = set()

        # Plot each observation
        for i, obs in enumerate(self.observations):
            # Time range and altitude calculations
            # TODO clean up this part by using existing variables in the obs objects
            time_range = Time(
                np.linspace(obs.start_time, (obs.start_time + obs.exposure_time), 20),
                format="jd",
            ).datetime
            altitudes = obs.target.coords.transform_to(
                night.observer.altaz(time=time_range)
            ).alt.deg
            # Generate an array of equally spaced Julian Dates
            num_points = 300  # The number of points you want
            jd_array = np.linspace(sunset.jd, sunrise.jd, num_points)

            # Convert back to Astropy Time objects
            night_time_array = Time(jd_array, format="jd", scale="utc").datetime

            night_altitudes = obs.target.coords.transform_to(
                night.observer.altaz(time=night_time_array)
            ).alt.deg

            # Determine whether to add to legend
            program_id = obs.target.program.progID
            add_to_legend = program_id not in added_to_legend
            if add_to_legend:
                added_to_legend.add(program_id)

            # Hover text
            hovertemplate = (
                f"<b>{obs.target.name}</b><br>"
                + f"{obs.target.program.instrument} {program_id}<br><br>"
                + f"Start time: {Time(obs.start_time, format='jd').datetime.strftime('%H:%M:%S %d-%m-%Y')}<br>"
                + f"Exp time: {timedelta(obs.exposure_time)}<br>"
                + f"Comment: {obs.target.comment}"
                + "<extra></extra>"
            )

            # Plotting the target
            fig.add_trace(
                go.Scatter(
                    x=night_time_array,
                    y=night_altitudes,
                    mode="lines",
                    line=dict(color="gray", dash="dot", width=0.3),
                    hoverinfo="skip",
                    showlegend=False,
                ),
                secondary_y=False,
            )
            fig.add_trace(
                go.Scatter(
                    x=time_range,
                    y=altitudes,
                    mode="lines",
                    line=dict(color=prog_colors[program_id], width=2),
                    name=f"{program_id}",
                    text=f"{obs.target.name}",
                    hovertemplate=hovertemplate,
                    # hovertext=hover_text,
                    # hoverinfo="text",
                    showlegend=add_to_legend,
                ),
                secondary_y=False,
            )

        # Twilight zones plotting
        shaded_dicts = {}
        shades = list(
            zip(
                [
                    sunset.datetime,
                    civil_evening,
                    nautical_evening,
                    astronomical_evening,
                    astronomical_morning,
                    nautical_morning,
                    civil_morning,
                    sunrise.datetime,
                ],
                ["", "yellow", "orange", "red", "", "red", "orange", "yellow"],
            )
        )
        # print(shades)
        for i, pair in enumerate(shades):
            if (i == 0) or (i == 4):
                continue
            twilight, color = pair
            shaded_dicts[twilight] = dict(
                type="rect",
                x0=shades[i - 1][0],
                y0=0,
                x1=twilight,
                y1=90,
                line=dict(width=0),
                fillcolor=color,
                opacity=0.2,
            )
        for twilight in shaded_dicts:
            fig.add_shape(**shaded_dicts[twilight])

        # Set up airmass values and compute the corresponding altitudes for those airmass values
        desired_airmasses = np.arange(1.8, 0.9, -0.1)
        corresponding_altitudes = list(
            90.0 - np.degrees(np.arccos(1.0 / desired_airmasses[:-1]))
        )
        corresponding_altitudes.append(90.0)

        # Formatting x-axis and y-axis
        fig.update_xaxes(title_text="Time [UTC]", tickformat="%H:%M", tickangle=45)
        fig.update_yaxes(title_text="Altitude [deg]", range=[30, 90], secondary_y=False)
        fig.update_yaxes(
            title_text="Airmass",
            tickvals=corresponding_altitudes,
            ticktext=np.round(desired_airmasses, 2),
            secondary_y=True,
        )

        # Legend and title
        fig.update_layout(
            title=f"Schedule for the night of {self.observations[0].night.night_date}",
            legend_title_text="Program IDs",
            legend=dict(orientation="v", xanchor="left", yanchor="top"),
            plot_bgcolor="white",
        )

        # Saving the plot if requested
        if path is not None:
            fig.write_image(path, format="png")

        fig.show()

    def plot_polar(self, display: bool = True, path: str = None):
        """
        Plot the azimuth vs. altitude of the observations in a polar plot.

        Parameters
        ----------
        display : bool, optional
            Option to display the plot. Defaults to True.
        path : str, optional
            The path to the file where the plot will be saved. Defaults to None.
        """
        # Create polar plot
        plt.figure(figsize=(8, 8))
        ax = plt.subplot(111, polar=True)
        for i, obs in enumerate(self.observations):
            # Plot each observation as a line in the polar plot
            ax.plot(
                np.radians(obs.obs_azimuths),
                obs.obs_altitudes,
                # label=obs.target.program.progID,
                color=obs.target.program.plot_color,
            )
            if i == 0:
                # Plot a greend point at the start of the night
                ax.plot(
                    np.radians(obs.obs_azimuths[0]),
                    obs.obs_altitudes[0],
                    "go",
                    label="Start of the plan",
                )
            elif i == len(self.observations) - 1:
                # Plot a red point at the end of the night
                ax.plot(
                    np.radians(obs.obs_azimuths[-1]),
                    obs.obs_altitudes[-1],
                    "ro",
                    label="End of the plan",
                )

            # Connect the last point of the current observation with the first point of the next observation
            if i < len(self.observations) - 1:
                ax.plot(
                    [
                        np.radians(obs.obs_azimuths[-1]),
                        np.radians(self.observations[i + 1].obs_azimuths[0]),
                    ],
                    [obs.obs_altitudes[-1], self.observations[i + 1].obs_altitudes[0]],
                    ls="dashed",
                    lw=0.6,
                    c="gray",
                )

        # Additional plot formatting
        ax.set_theta_zero_location("N")  # Set 0 degrees at the top
        ax.set_theta_direction(-1)  # Set the rotation of the plot clockwise
        ax.set_title("Azimuth vs. Altitude", va="bottom")
        ax.set_ylim(90, 30)
        plt.legend(loc="best")

        # Saving the plot if requested
        if path is not None:
            plt.tight_layout()
            plt.savefig(path, format="png")

        if display:
            plt.show()
        else:
            plt.close()

    def _instruments_analysis(self):
        """
        Analyze the instruments used in the plan. It registers which instruments are used in this
        plan and assigns a default plot color if the user didn't provide one.
        """
        self.instruments: List[Instrument] = []
        for obs in self.observations:
            instrument = obs.target.program.instrument
            if instrument not in self.instruments:
                self.instruments.append(instrument)

        # Assign default colors to instruments that don't have a color assigned
        default_colors = itertools.cycle(
            [mcolors.rgb2hex(color) for color in plt.get_cmap("Pastel1").colors]
        )
        for instrument in self.instruments:
            if instrument.plot_color is None:
                instrument.plot_color = next(default_colors)

    def plot_altaz(self, display: bool = True, path: str = None) -> None:
        """
        Plot the azimuth and altitude of the observations.

        Parameters
        ----------
        display : bool, optional
            Option to display the plot. Defaults to True.
        path : str, optional
            The path to the file where the plot will be saved. Defaults to None.
        """
        # Initialize the figure and subplots
        _, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 8), sharex=True)

        # Analyse the instruments in this plan
        self._instruments_analysis()

        # Iterate through observations and plot them
        for i, obs in enumerate(self.observations):
            # Create arrays of JD for each observation point
            jd_times = np.linspace(obs.start_time, obs.end_time, len(obs.obs_azimuths))

            # TODO if a line crosses the 0/360 deg boundary create a split right at the boundary
            # and plot the two parts separately. This is so that the line is not drawn across the
            # entire plot when switching sides.
            # Plot azimuths and altitudes
            ax1.plot(jd_times, obs.obs_azimuths, "-", lw=1, c="k")
            ax2.plot(jd_times, obs.obs_altitudes, "-", lw=1, c="k")

            # Plot diagonal line between this obs and start of the next
            if i < len(self.observations) - 1:
                ax1.plot(
                    [jd_times[-1], self.observations[i + 1].start_time],
                    [obs.obs_azimuths[-1], self.observations[i + 1].obs_azimuths[0]],
                    "-"
                    if abs(
                        obs.obs_azimuths[-1] - self.observations[i + 1].obs_azimuths[0]
                    )
                    <= 180
                    else "--",
                    lw=1,
                    c="k",
                )
                ax2.plot(
                    [jd_times[-1], self.observations[i + 1].start_time],
                    [obs.obs_altitudes[-1], self.observations[i + 1].obs_altitudes[0]],
                    "-",
                    lw=1,
                    c="k",
                )

        # Shaded regions
        for i in range(len(self.observations)):
            obs = self.observations[i]
            # Shade the observation time color coded by instrument
            ax1.axvspan(
                obs.start_time,
                obs.end_time,
                color=obs.target.program.instrument.plot_color,
                alpha=0.7,
            )
            ax2.axvspan(
                obs.start_time,
                obs.end_time,
                color=obs.target.program.instrument.plot_color,
                alpha=0.7,
            )
            if i < len(self.observations) - 1:
                # Shade the overhead between observations
                end_current_obs = obs.end_time
                start_next_obs = self.observations[i + 1].start_time
                if start_next_obs > end_current_obs:  # Ensure there is a gap
                    ax1.axvspan(
                        end_current_obs, start_next_obs, color="grey", alpha=0.8
                    )
                    ax2.axvspan(
                        end_current_obs, start_next_obs, color="grey", alpha=0.8
                    )

        # Use scatter plot for the legend markers
        for inst in self.instruments:
            ax1.scatter(
                [],
                [],
                label=inst.name,
                color=inst.plot_color,
                edgecolor="none",
                s=100,
                marker="s",
            )

        # Adjusting legend
        ax1.legend(loc="best", title="Instruments")

        # Set the labels and titles
        ax1.set_ylabel("Azimuth Angle [deg]")
        ax1.set_ylim(0, 360)
        ax2.set_ylabel("Elevation Angle [deg]")
        ax2.set_ylim(30, 90)
        ax2.set_xlabel("Observation Time [JD]")

        # Set titles and layout
        ax1.set_title(
            "Azimuth\n(dashed line means it actually crosses 0/360 deg and goes the other way than shown)"
        )
        ax2.set_title("Elevation")

        # Saving the plot if requested
        if path is not None:
            plt.tight_layout()
            plt.savefig(path, format="png")

        if display:
            plt.show()
        else:
            plt.close()

    def to_df(self):
        """
        Create a pandas DataFrame of the plan and return it. It saves this table in an attribute
        called 'df'.
        """
        # List to hold the rows for the DataFrame
        data = []

        def format_timedelta(tdelta):
            """Format timedelta to display only hours, minutes, and seconds."""
            total_seconds = int(tdelta.total_seconds())
            hours, remainder = divmod(total_seconds, 3600)
            minutes, seconds = divmod(remainder, 60)
            return f"{hours:02}:{minutes:02}:{seconds:02}"

        for obs in self.observations:
            start_time = Time(obs.start_time, format="jd").datetime
            exp_time = timedelta(days=obs.exposure_time)
            ra_str = obs.target.coords.ra.to_string(
                unit="hour", sep=":", precision=0, pad=True
            )
            dec_str = obs.target.coords.dec.to_string(
                unit="deg", sep=":", precision=0, pad=True, alwayssign=True
            )

            # Add a dictionary for each observation
            data.append(
                {
                    "Instrument": obs.target.program.instrument,
                    "ProgID": obs.target.program.progID,
                    "Target": obs.target.name,
                    "RA": ra_str,
                    "DEC": dec_str,
                    "Start Time": start_time.strftime("%H:%M:%S"),
                    "Exp Time": format_timedelta(exp_time),
                    "Exp Time (s)": int(exp_time.total_seconds()),
                    "Comment": obs.target.comment,
                }
            )

        # Create the DataFrame from the list of dictionaries
        columns = [
            "Instrument",
            "ProgID",
            "Target",
            "RA",
            "DEC",
            "Start Time",
            "Exp Time",
            "Exp Time (s)",
            "Comment",
        ]
        self.df = pd.DataFrame(data, columns=columns)

        return self.df

    def to_csv(self, *args, **kwargs):
        """
        Create a CSV file for the pandas DataFrame representation of the Plan.
        This method can take any keyword argument that the pd.DataFrame.to_csv() method can take.

        Parameters
        ----------
        *args
            Positional arguments that will be passed to pd.DataFrame.to_csv()
        **kwargs
            Keyword arguments that will be passed to pd.DataFrame.to_csv()
        """
        if not hasattr(self, "df"):
            self.to_df()
        self.df.to_csv(*args, **kwargs)

    def __len__(self):
        return len(self.observations)

    def __repr__(self):
        return f"<Plan> with {len(self)} observations"

    def __str__(self):
        return self.to_df().to_string()



from datetime import date

import pytest
from astroplan import Observer
from astropy.coordinates import Angle, SkyCoord

from scopes.merits import airmass
from scopes.scheduler_components import (
    Instrument,
    Merit,
    Night,
    Observation,
    Overheads,
    Plan,
    Program,
    Target,
)


@pytest.fixture
def example_observer():
    return Observer.at_site("lasilla")


@pytest.fixture
def example_night(example_observer):
    return Night(
        night_date=date(2024, 8, 1),  # National Day of Switzerland
        observations_within="astronomical",
        observer=example_observer,
    )


@pytest.fixture
def example_instrument():
    return Instrument(
        name="Example Spectrograph",
        instrument_type="Spectrograph",
        plot_color="#FF5733",
    )


@pytest.fixture
def example_program(example_instrument):
    return Program(
        progID="Prog123",
        instrument=example_instrument,
        priority=1,
        time_share_allocated=0.3,
    )


@pytest.fixture
def example_merit():
    return Merit(
        name="Example Merit",
        func=airmass,
        merit_type="veto",
        parameters={"limit": 1.8},
    )


@pytest.fixture
def example_custom_merit_func():
    def custom_merit_function(observation, example_param):
        """Example merit function."""
        return 1.0

    return custom_merit_function


@pytest.fixture
def example_coords():
    # A coordinate in the sky for a hypothetical target that is visible during the night
    # of the example_night fixture which is 1st August 2024
    return SkyCoord(ra="20h00m00s", dec="-20d00m00s", frame="icrs")


@pytest.fixture
def example_target(example_program, example_coords):
    target = Target(
        name="Example Target",
        program=example_program,
        coords=example_coords,
        priority=1,
        comment="This is a test target",
    )
    return target


@pytest.fixture
def example_fairness_merit():
    def fairness_merit_func(observation):
        return 1.2  # Example fairness merit score

    return Merit(
        name="Fairness Merit",
        func=fairness_merit_func,
        merit_type="fairness",
        parameters={},
    )


@pytest.fixture
def example_efficiency_merit():
    def efficiency_merit_func(observation):
        return 0.9  # Example efficiency merit score

    return Merit(
        name="Efficiency Merit",
        func=efficiency_merit_func,
        merit_type="efficiency",
        parameters={},
    )


@pytest.fixture
def example_veto_merit():
    def veto_merit_func(observation):
        return 0.0  # Example veto merit score (vetoes the observation)

    return Merit(
        name="Veto Merit", func=veto_merit_func, merit_type="veto", parameters={}
    )


@pytest.fixture
def example_observation(example_target, example_merit):
    example_target.add_merit(merit=example_merit)
    return Observation(target=example_target, exposure_time=600.0)


@pytest.fixture
def example_observation_with_fairness(example_target, example_fairness_merit):
    example_target.add_merit(merit=example_fairness_merit)
    return Observation(target=example_target, exposure_time=600.0)


@pytest.fixture
def example_observation_with_efficiency(example_target, example_efficiency_merit):
    example_target.add_merit(merit=example_efficiency_merit)
    return Observation(target=example_target, exposure_time=600.0)


@pytest.fixture
def example_observation_with_veto(example_target, example_veto_merit):
    example_target.add_merit(merit=example_veto_merit)
    return Observation(target=example_target, exposure_time=600.0)


@pytest.fixture
def example_full_observation1(example_observation: Observation, example_night: Night):
    example_observation.set_night(example_night)
    example_observation.set_start_time(2460524.7)
    example_observation.skypath()
    example_observation.update_start_and_score(2460524.7)
    return example_observation


@pytest.fixture
def example_full_observation2(
    example_coords: SkyCoord, example_night: Night, example_program, example_merit
):
    new_coords = example_coords.directional_offset_by(Angle("1d"), Angle("1d"))
    new_target = Target(
        name="Target002", program=example_program, coords=new_coords, priority=1
    )
    new_target.add_merit(merit=example_merit)
    example_observation2 = Observation(target=new_target, exposure_time=600.0)
    example_observation2.set_night(example_night)
    example_observation2.set_start_time(2460524.7)
    example_observation2.skypath()
    example_observation2.update_start_and_score(2460524.708)
    return example_observation2


@pytest.fixture
def example_full_observation3(
    example_coords: SkyCoord, example_night: Night, example_program, example_merit
):
    new_coords = example_coords.directional_offset_by(Angle("2d"), Angle("2d"))
    new_target = Target(
        name="Target003", program=example_program, coords=new_coords, priority=1
    )
    new_target.add_merit(merit=example_merit)
    example_observation3 = Observation(target=new_target, exposure_time=600.0)
    example_observation3.set_night(example_night)
    example_observation3.set_start_time(2460524.7)
    example_observation3.skypath()
    example_observation3.update_start_and_score(2460524.716)
    return example_observation3


@pytest.fixture
def example_overheads():
    return Overheads(slew_rate_az=1.0, slew_rate_alt=1.0, cable_wrap_angle=180.0)


@pytest.fixture
def example_plan(
    example_full_observation1, example_full_observation2, example_full_observation3
):
    plan = Plan()
    plan.add_observation(example_full_observation1)
    plan.add_observation(example_full_observation2)
    plan.add_observation(example_full_observation3)
    return plan



import pytest

from scopes.scheduler_components import Instrument


def test_instrument_initialization():
    instrument = Instrument(
        name="Spectrograph", instrument_type="Spectrograph", plot_color="#FF5733"
    )
    assert instrument.name == "Spectrograph"
    assert instrument.instrument_type == "Spectrograph"
    assert instrument.plot_color == "#FF5733"


def test_instrument_default_initialization():
    instrument = Instrument(name="Camera")
    assert instrument.name == "Camera"
    assert instrument.instrument_type == ""
    assert instrument.plot_color is None


def test_instrument_name_type_error():
    with pytest.raises(TypeError):
        Instrument(name=123)  # Name must be a string


def test_instrument_type_type_error():
    with pytest.raises(TypeError):
        Instrument(name="Spectrograph", instrument_type=123)  # Type must be a string


def test_instrument_invalid_plot_color():
    with pytest.raises(ValueError):
        Instrument(name="Spectrograph", plot_color="UH0095")


def test_instrument_none_plot_color():
    instrument = Instrument(name="Spectrograph", plot_color=None)
    assert instrument.plot_color is None


def test_instrument_str_method():
    instrument = Instrument(name="Spectrograph")
    assert str(instrument) == "Spectrograph"


def test_instrument_repr_method_with_type_and_color():
    instrument = Instrument(
        name="Spectrograph", instrument_type="Spectrograph", plot_color="#FF5733"
    )
    expected_repr = "Instrument(\n    Name = Spectrograph\n    Type = Spectrograph\n    Plot color = #FF5733"
    assert repr(instrument).startswith(expected_repr)


def test_instrument_repr_method_with_type():
    instrument = Instrument(name="Spectrograph", instrument_type="Spectrograph")
    expected_repr = "Instrument(\n    Name = Spectrograph\n    Type = Spectrograph"
    assert repr(instrument).startswith(expected_repr)


def test_instrument_repr_method_without_type_and_color():
    instrument = Instrument(name="Spectrograph")
    expected_repr = "Instrument(\n    Name = Spectrograph"
    assert repr(instrument).startswith(expected_repr)


def test_instrument_equality():
    instrument1 = Instrument(name="Spectrograph", instrument_type="Spectrograph")
    instrument2 = Instrument(name="Spectrograph", instrument_type="Camera")
    instrument3 = Instrument(name="Camera", instrument_type="Spectrograph")
    assert instrument1 == instrument2  # Same name, should be equal
    assert instrument1 != instrument3  # Different names, should not be equal
    assert instrument1 != "Not an Instrument"  # Should not be equal to a different type



from datetime import date

import pytest
from astroplan import Observer
from astropy.time import Time

from scopes.scheduler_components import Night


def test_night_initialization(example_night):
    assert example_night.night_date == date(2024, 8, 1)
    assert example_night.observations_within == "astronomical"
    assert isinstance(example_night.observer, Observer)


def test_invalid_observations_within(example_observer):
    with pytest.raises(ValueError):
        Night(
            night_date=date(2024, 8, 1),
            observations_within="invalid",
            observer=example_observer,
        )


def test_solar_midnight_calculation(example_night):
    assert isinstance(example_night.solar_midnight, Time)


def test_twilight_times(example_night):
    assert isinstance(example_night.sunset, Time)
    assert isinstance(example_night.sunrise, Time)
    assert isinstance(example_night.civil_evening, float)
    assert isinstance(example_night.nautical_evening, float)
    assert isinstance(example_night.astronomical_evening, float)
    assert isinstance(example_night.civil_morning, float)
    assert isinstance(example_night.nautical_morning, float)
    assert isinstance(example_night.astronomical_morning, float)


def test_night_time_range(example_night):
    assert len(example_night.night_time_range) == 300


# def test_culmination_window(example_night):
#     TODO: Create test Observations
#     obs1 = Observation()
#     obs2 = Observation()

#     obs_list = [obs1, obs2]
#     example_night.calculate_culmination_window(obs_list)
#     assert example_night.culmination_window == (2456789.5, 2456789.5)


def test_str_representation(example_night):
    assert str(example_night).startswith("Night(Date: 2024-08-01,")


def test_night_equality(example_night):
    other_night = Night(
        night_date=date(2024, 8, 1),
        observations_within="astronomical",
        observer=example_night.observer,
    )
    assert example_night == other_night
    different_night = Night(
        night_date=date(2024, 8, 2),
        observations_within="astronomical",
        observer=example_night.observer,
    )
    assert example_night != different_night



"""Definition of all the generic merit functions"""

import warnings

import numpy as np

# from astropy.time import Time
from .scheduler_components import Observation


def time_share(
    observation: Observation, alpha: int = 3, beta: float = 5.0, delta: float = 0.05
) -> float:
    """
    Time share fairness merit. It uses a modified sigmoid function to calculate the merit.
    The specific shape can be set with the parameters alpha and beta. The exact formula is:

    m = (delta / (1 + np.exp((pct_diff / beta) ** alpha))) + (1 - delta / 2)

    It's shaped in a way so that there is some permissiveness. This means that a program
    can be over or under the allocated time by a certain percentage before its priority is decreased
    or increased. The alpha parameter controls how sharp the sigmoid is, and the beta parameter
    controls how much difference is allowed (in percentage).

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    alpha : float, optional
        The attack parameter for how sharp the sigmoid is. Defaults to 3.
    beta : float, optional
        The leeway parameter for how much difference in time use is allowed. Defaults to 5.
    delta : float, optional
        The maximum percentage increase or decrease that will be applied if a program is over or
        under its allocated time. Defaults to 0.1.

    Returns
    -------
    merit : float
        The time share merit of the observation
    """
    # Check that the parameters are valid
    if alpha <= 0:
        raise ValueError("alpha for time_share merit must be greater than 0")
    if alpha % 2 == 0:
        raise ValueError("alpha for time_share merit must be an odd positive integer")
    if beta <= 0.0:
        raise ValueError("beta for time_share merit must be greater than 0")
    if delta <= 0.0:
        raise ValueError("delta for time_share merit must be greater than 0")

    # Calculate the time share of the observation
    pct_diff = observation.target.program.time_share_pct_diff * 100
    exp_term = (pct_diff / beta) ** alpha

    # If the exponent term is too big, cap it at 5 or -5
    # This is to limit the size of the exponent term and control the np.exp() function.
    # After 5 the merit already reaches its limits
    if abs_exp_term := abs(exp_term) > 5:
        sign = exp_term / abs_exp_term
        exp_term = sign * 5

    # Calculate merit
    merit = (delta / (1 + np.exp(exp_term))) + (1 - delta / 2)
    return merit


def priority(
    observation: Observation,
    prog_base: float = 1.0,
    prog_offset: float = 0.1,
    tar_base: float = 0.0,
    tar_offset: float = 0.05,
) -> float:
    """
    Priority merit function. This is a fairness merit function to schedule based on priority.
    In this implementation it assumes that the priority values given is between 0 and 3, where
    0 is the highest priority and 3 is the lowest. This priority value is mapped to a base values
    for the priority level of 2, and then an offset is added above or below that base value depending
    if the priority is 3, 1, or 0. This is to convert this priority scale to something close to
    0 or 1 usually.

    The default values reflect a system where the program priorities are centered at 1, and then
    the target priorities are centered at 0 which work as a modifier to the program priority. This
    means that the priority of the program takes precedent, and then the priority of the targets
    help to decide within the program which targets to observe first. There is a small overlap,
    where a target with priority 0 can have a mapped priority higher than a target with priority 3
    of the next highest priority program. At this stage we don't have to worry that this generates
    more observations for the higher priority programs because that is balanced by the time share
    merit.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    prog_base : float
        The base value for the priority level of 2, for programs.
    prog_offset : float
        The offset to be added to the base value of programs for the priority levels of 3, 1 and 0.
    tar_base : float
        The base value for the priority level of 2, for targets.
    tar_offset : float
        The offset to be added to the base value of targets for the priority levels of 3, 1 and 0.
    """
    # Validate priority value for program and target exists
    if observation.target.program.priority is None:
        raise ValueError("Program priority value is None")
    if observation.target.priority is None:
        raise ValueError("Target priority value is None")

    map_prog_priority = prog_base + prog_offset * (
        2 - observation.target.program.priority
    )
    map_tar_priority = tar_base + tar_offset * (2 - observation.target.priority)
    return map_prog_priority + map_tar_priority


def at_night(observation: Observation) -> float:
    """
    Merit function that returns 1 if the observation is within the chosen night time limits, and
    0 otherwise.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    """
    return float(
        observation.start_time >= observation.night.obs_within_limits[0]
        and observation.end_time <= observation.night.obs_within_limits[1]
    )


def airmass(observation: Observation, limit: float, alpha: float = 0.0001) -> float:
    """
    Merit function on the current airmass of the target. It uses a hyperbolic tangent function to
    gradually increase the merit as the airmass decreases. The specific shape can be set with the
    parameters limit and alpha. The airmass considered is the maximum that the observation reaches
    during its exposure time. The exact formula is:

    m = (tanh((limit - current_airmass) / alpha) + 1) / 2

    By default the alpha parameter is set to 0.0001 which in practice is equivalent to a step
    function, or a simple boolean with a hard limit.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    limit : float, optional
        The limit airmass.
    alpha : float, optional
        A measure of the tolerance around the maximum airmass. Defaults to 0.0001, equivalent to a
        step function.
    """
    if len(observation.obs_airmasses) == 0:
        return 0.0
    else:
        # return observation.obs_airmasses.max() < limit
        arg = (limit - np.max(observation.obs_airmasses)) / alpha
        return (np.tanh(arg) + 1) / 2


def altitude(
    observation: Observation,
    min: float = 20,
    max: float = 90,
) -> float:
    """
    Altitude merit function.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    """
    # Claculate altitude throughout the exposure of the observation
    if len(observation.obs_altitudes) == 0:
        return 0.0
    else:
        return (observation.obs_altitudes.min() > min) and (
            observation.obs_altitudes.max() < max
        )


def moon_separation(
    observation: Observation,
    theta_lim: float = 20.0,
    theta_start: float = 30.0,
    alpha: float = 3.0,
) -> float:
    """
    Moon separation constraint merit function
    TODO: this entire merit function has to be tested.

    m = ((sep - min_sep) / (max_sep - min_sep)) ** alpha

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    min : float, optional
        The minimum separation to the moon in degrees. Defaults to 30.0.
    """
    if alpha <= 0:
        raise ValueError("alpha for moon_separation merit must be greater than 0")
    if theta_lim >= theta_start:
        raise ValueError("theta_lim must be less than theta_start")
    if theta_lim <= 0 or theta_start <= 0:
        raise ValueError("theta_lim and theta_start must be greater than 0")
    # Find the moon altaz position at the times of the observation
    moon_altaz_frame = observation.night.moon_altaz_frame
    start_idx = np.searchsorted(
        moon_altaz_frame.obstime.jd, observation.start_time, side="left"
    )
    end_idx = np.searchsorted(
        moon_altaz_frame.obstime.jd, observation.end_time, side="right"
    )
    # Calculate moon distance throughout the exposure of the observation
    sep = np.min(
        observation.target.coords.separation(moon_altaz_frame[start_idx:end_idx]).deg
    )

    # Calculate the value of the merit
    if sep < theta_lim:
        return 0.0
    elif sep >= theta_start:
        return 1.0
    else:
        return ((sep - theta_lim) / (theta_start - theta_lim)) ** alpha


# TODO: Implement a merit that takes into account the moon illumination and distance
# def moon_illumination(observation: Observation, min: float = 0.0) -> float:
#     """
#     Moon illumination constraint merit function

#     Parameters
#     ----------
#     observation : Observation
#         The Observation object to be used
#     min : float, optional
#         The minimum moon illumination. Defaults to 0.0.
#     """
#     # Find the moon altaz position at the times of the observation
#     moon_illumination = observation.night.moon_illumination
#     start_idx = np.searchsorted(moon_illumination.obstime.jd, observation.start_time, side="left")
#     end_idx = np.searchsorted(moon_illumination.obstime.jd, observation.end_time, side="right")
#     # Calculate moon distance throughout the exposure of the observation
#     illum = moon_illumination[start_idx:end_idx]

#     if illum.min() < min:
#         return 0.0
#     else:
#         return 1.0


def end_time(observation: Observation, time: float) -> float:
    """
    Veto merit function that returns 0 if the observation ends after the given time, and 1
    otherwise.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    time : float
        The end time of the observation
    """
    return float(observation.end_time <= time)


def start_time(observation: Observation, time: float) -> float:
    """
    Veto merit function that returns 0 if the observation starts before the given time, and 1
    otherwise.

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    time : float
        The start time of the observation
    """
    return float(observation.start_time >= time)


def culmination(observation: Observation, verbose: bool = False) -> float:
    """
    Culmination constraint merit function.
    This merit calculates the current height of the target and the proportion to the maximum height
    it will reach during the current night.
    """
    # Calculate the current altitude of the target
    current_altitude = observation.obs_altitudes[0]
    # Calculate altitude proportional to available altitue range
    altitude_prop = (current_altitude - observation.min_altitude) / (
        observation.max_altitude - observation.min_altitude
    )
    if verbose:
        print(f"Current altitude: {current_altitude}")
        print(f"Max altitude: {observation.max_altitude}")
        print(f"Min altitude: {observation.min_altitude}")
        print(f"Altitude proportion: {altitude_prop}")
    return altitude_prop


def culmination_efficiency(observation: Observation, verbose: bool = False) -> float:
    """
    This merit is designed to make the scheduling of astronomical observations more efficient by
    expanding the selection of observable stars beyond those reaching their culmination (highest
    point in the sky) during the night. The normal Culmination merit misses out on stars that
    culminate before the night begins or after it ends, even though these stars are still
    observable at acceptable altitudes during the night.

    To address this, it uses an extended time range for calculating merits beyond the actual
    observable hours of the night. This extended range includes the earliest culmination point of
    any star still observable at the night's start and the latest culmination point of any star
    still observable at the night's end.

    This extended timerange is then mapped to the actual night time range where observations will
    be taken (typically within nautical or astronomical twilights). Its a simple one-to-one mapping
    between two different time ranges. Then to calcualte the merit, the time at which the star
    actually culminates is mapped from the larger timerange to the actual night range. That new
    mapped time is when that star will peak in this merit function.

    The method shifts observation priorities throughout the night—prioritizing setting stars early
    on, stars near their culmination around the middle, and rising stars towards the end. This
    method is a compromise between observing stars at their highest point but also giving priority
    to stars that are setting or rising.


    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    verbose : bool, optional
        If True, print the calculated merit. Defaults to False.
    """
    n = observation.night  # Simplify the notation for shorter and cleaner code
    # Check if the culmination window is within the night time limits
    # and adjust the lower and upper limits of the window if necessary
    # This is to make sure that if the culmination window is within the full night time limits
    # then the mapped window will be the same as the original window. Otherwise it would map
    # from a smaller window to a larger window, which would be wrong.
    mapped_window = [n.obs_within_limits[0], n.obs_within_limits[1]]
    if n.culmination_window[0] > n.obs_within_limits[0]:
        mapped_window[0] = n.culmination_window[0]
    if n.culmination_window[1] < n.obs_within_limits[1]:
        mapped_window[1] = n.culmination_window[1]

    time_prop = (observation.culmination_time - n.culmination_window[0]) / (
        n.culmination_window[1] - n.culmination_window[0]
    )
    peak_merit_time = n.obs_within_limits[0] + time_prop * (
        n.obs_within_limits[1] - n.obs_within_limits[0]
    )

    # Calculate the merit of the target at the mapping time
    merit = gaussian((peak_merit_time - observation.start_time), 4 / 24)

    if verbose:
        print(f"{time_prop = }")
        print(f"{peak_merit_time = }")
        print(f"{merit = }")

    return merit


# def periodic_gaussian(
#     observation: Observation,
#     epoch: float,
#     period: float,
#     sigma: float,
#     phases: list = [0.0],
#     verbose: bool = False,
# ) -> float:
#     """
#     # TODO Change this entire merit with the version I show in the documentation as its more stable
#     and the sigma actually makes physical sense.

#     Periodic Gaussian merit function.

#     Analytic expression: exp(-0.5(sin(2pi(x-epoch)/period)/s)^2)

#     The introduction of the sine function breaks the traditional meaning of the standard deviation.
#     So the s parameter will have to be finetuned depending on the period.

#     Parameters
#     ----------
#     x : float
#         The x value at which to evaluate the merit function.
#     epoch : float
#         Where the peak of the merit will be centered
#     period : float
#         The period of the Gaussian.
#     sigma : float
#         Measure of the width of each Gaussian.
#     phases : list, optional
#         List of phases at which the gaussians should peak. Defaults to [0.0].
#     verbose : bool, optional
#         If True, print the calculated merit. Defaults to False.
#     """
#     merit = 0.0
#     for phase in phases:
#         merit += np.exp(
#             -0.5
#             * (
#                 np.sin(
#                     np.pi * (observation.start_time - (epoch + phase * period)) / period
#                 )
#                 / sigma
#             )
#             ** 2
#         )
#     if verbose:
#         print(f"current phase = {((observation.start_time - epoch) % period) / period}")
#         print(f"{merit = }")
#     return merit


def phase_specific(
    observation: Observation,
    epoch: float,
    period: float,
    sigma: float,
    phases: list = [0.0],
) -> float:
    """
    This merit is used when the observation should be taken at specific phases of a periodic
    time interval. In exoplanet detections, for example, this is used when observations have
    to be taken at a specific phase of the orbit of a planet. Its analytic expression is a slight
    modification of the merit presented by Granzer (2004):

    .. math::

        m(x, \\phi) = \\sum_{i=-1}^{1} \\exp\\left(-\\frac{(x(t) - (\\phi - i))^2}{2 \\sigma^2}\\right)

    .. math::

        x(t) = \\frac{\\mod(t - t0, p)}{p}, \\text{ for } |x| \\leq 0.5

    where t is time, p is the period, \\phi is the desired phase at which to observe, and \\sigma is the
    standard deviation of the Gaussian in phase space. If more than one phase is given, the merit
    will be the max of the merits for each phase:

    .. math::

        m = \\max_{\\phi} m(x, \\phi)

    Parameters
    ----------
    observation : Observation
        The Observation object to be used
    epoch : float
        The phase at which the merit will be centered
    period : float
        The period of the observation
    phases : list, optional
        List of phases at which the gaussians should peak. Defaults to [0.0].
    """
    # Value checks
    if sigma <= 0:
        raise ValueError("sigma for phase_specific merit must be greater than 0")
    if sigma > 0.2:
        raise warnings.warn(
            "Care should be taken when sigma > 0.2, as the merit can end up being \
                            significantly larger than 1."
        )
    if period <= 0:
        raise ValueError("period for phase_specific merit must be greater than 0")
    if np.any(np.array(phases) < 0) or np.any(np.array(phases) > 1):
        raise ValueError("phases for phase_specific merit must be between 0 and 1")

    # Calculate the merit for each phase
    merits = []
    for phase in phases:
        x = np.mod(observation.start_time - epoch, period) / period
        merit = 0
        for i in range(-1, 2):
            merit += np.exp(-0.5 * (((x - (phase - i)) / (sigma)) ** 2))
        merits.append(merit)

    return np.max(merits)


def time_critical(
    observation: Observation,
    start_time: float,
    start_time_tolerance: float,
    steepness: float = 0.0014,  # in days, which is ~2 minutes
    verbose: bool = False,
) -> float:
    """
    Calculate the time criticality merit of an observation. It uses a double hyperbolic tangent
    function to gradually increase the merit as the observation approaches the desired start time.
    The center times of the increasing tanh and decresing tanh, are start_time - start_time_tolerance
    and start_time + start_time_tolerance, respectively.

    TODO Rethink this merit and how to ensure that time critical observations are done without
    failure. This merit is not enough to ensure that the observation is done at the desired time
    as a previous observation can cover the entire tolerance range and thus block the time critical
    observation from being done.

    Parameters
    ----------
    observation : Observation
        The observation object.
    start_time : float
        The desired start time for the observation. In Julian Date (JD).
    start_time_tolerance : float
        The tolerance around the desired start time in days.
    steepness : float, optional
        The steepness of the hyperbolic tangent function. A measure of how much time it takes the
        function to go from 0 to the max value in days. Defaults to 0.0014 days, which is ~2 minutes.
    verbose : bool, optional
        If True, print the calculated merit. Defaults to False.

    Returns
    -------
    float: The time criticality merit of the observation.
    """
    arg1 = (observation.start_time - (start_time - start_time_tolerance)) / steepness
    arg2 = ((start_time + start_time_tolerance) - observation.start_time) / steepness
    merit = np.tanh(arg1) + np.tanh(arg2)

    if verbose:
        print(f"time_critical {merit = }")
    return merit


def airmass_efficiency(observation: Observation) -> float:
    """
    Airmass efficiency merit function. Defined as the inverse of the maximum airmass reached during
    the observation.

    Parameters
    ----------
    observation : Observation
        The observation object.
    """
    if len(observation.obs_airmasses) == 0:
        return 0.0
    else:
        return 1 / np.max(observation.obs_airmasses)


def gaussian(x, sigma):
    """
    A simple Gaussian.

    Parameters
    ----------
    x : float
        The x value at which to evaluate the Gaussian.
    sigma : float
        Measure of the width of the Gaussian.
    """
    return np.exp(-0.5 * (x / sigma) ** 2)



import pytest

from scopes.scheduler_components import Program


def test_program_initialization(example_instrument):
    program = Program(
        progID="Prog123",
        instrument=example_instrument,
        priority=1,
        time_share_allocated=0.5,
        plot_color="#00FF00",
    )
    assert program.progID == "Prog123"
    assert program.instrument == example_instrument
    assert program.priority == 1
    assert program.time_share_allocated == 0.5
    assert program.plot_color == "#00FF00"


def test_program_default_initialization(example_instrument):
    program = Program(progID="Prog123", instrument=example_instrument, priority=2)
    assert program.progID == "Prog123"
    assert program.instrument == example_instrument
    assert program.priority == 2
    assert program.time_share_allocated == 0.0
    assert program.plot_color is None


def test_program_invalid_progID_type(example_instrument):
    with pytest.raises(TypeError):
        Program(
            progID=123, instrument=example_instrument, priority=1
        )  # progID must be a string


def test_program_invalid_instrument_type():
    with pytest.raises(TypeError):
        Program(
            progID="Prog123", instrument="NotAnInstrument", priority=1
        )  # Instrument must be of type Instrument


def test_program_invalid_priority_value(example_instrument):
    with pytest.raises(ValueError):
        Program(
            progID="Prog123", instrument=example_instrument, priority=5
        )  # Priority must be between 0 and 3
    with pytest.raises(ValueError):
        Program(
            progID="Prog123", instrument=example_instrument, priority=-1
        )  # Priority must be between 0 and 3


def test_program_invalid_priority_type(example_instrument):
    with pytest.raises(TypeError):
        Program(
            progID="Prog123", instrument=example_instrument, priority="High"
        )  # Priority must be an integer


def test_program_invalid_plot_color(example_instrument):
    with pytest.raises(ValueError):
        Program(
            progID="Prog123",
            instrument=example_instrument,
            priority=1,
            plot_color="UX0095",
        )


def test_program_invalid_time_share(example_instrument):
    # Time share must be between 0 and 1
    with pytest.raises(ValueError):
        Program(
            progID="Prog123",
            instrument=example_instrument,
            priority=1,
            time_share_allocated=1.5,
        )
    with pytest.raises(ValueError):
        Program(
            progID="Prog123",
            instrument=example_instrument,
            priority=1,
            time_share_allocated=-0.5,
        )


def test_set_current_time_usage_valid(example_instrument):
    program = Program(
        progID="Prog123",
        instrument=example_instrument,
        priority=1,
        time_share_allocated=0.4,
    )
    program.set_current_time_usage(0.5)
    assert program.time_share_current == 0.5
    assert (program.time_share_pct_diff - 0.1) < 1e-6


def test_set_current_time_usage_invalid(example_instrument):
    # current_time_usage must be between 0 and 1
    program = Program(progID="Prog123", instrument=example_instrument, priority=1)
    with pytest.raises(ValueError):
        program.set_current_time_usage(1.5)
    with pytest.raises(ValueError):
        program.set_current_time_usage(-0.5)


def test_program_str_method(example_instrument):
    program = Program(progID="Prog123", instrument=example_instrument, priority=1)
    expected_str = "Program(\n    ID = Prog123\n    Instrument = Example Spectrograph\n    Time allocated = 0.0\n    Priority = 1)"
    assert str(program) == expected_str


def test_program_repr_method(example_instrument):
    program = Program(progID="Prog123", instrument=example_instrument, priority=1)
    expected_repr = "Program(\n    ID = Prog123\n    Instrument = Example Spectrograph\n    Time allocated = 0.0\n    Priority = 1)"
    assert repr(program) == expected_repr



import pytest

from scopes.scheduler_components import Target


def test_target_initialization(example_target: Target, example_program, example_coords):
    assert example_target.name == "Example Target"
    assert example_target.program == example_program
    assert example_target.coords == example_coords
    assert example_target.ra_deg == example_coords.ra.deg
    assert example_target.dec_deg == example_coords.dec.deg
    assert example_target.priority == 1
    assert example_target.comment == "This is a test target"


def test_target_default_initialization(example_program, example_coords):
    target = Target(
        name="Example Target", program=example_program, coords=example_coords
    )
    assert target.name == "Example Target"
    assert target.program == example_program
    assert target.coords == example_coords
    assert target.priority is None
    assert target.comment == ""


def test_target_invalid_name_type(example_program, example_coords):
    with pytest.raises(TypeError):
        Target(
            name=123, program=example_program, coords=example_coords
        )  # Name must be a string


def test_target_invalid_program_type(example_coords):
    with pytest.raises(TypeError):
        Target(
            name="Example Target", program="InvalidProgram", coords=example_coords
        )  # Program must be of type Program


def test_target_invalid_coords_type(example_program):
    with pytest.raises(TypeError):
        Target(
            name="Example Target", program=example_program, coords="InvalidCoords"
        )  # Coords must be of type SkyCoord


def test_target_invalid_priority_value(example_program, example_coords):
    with pytest.raises(ValueError):
        Target(
            name="Example Target",
            program=example_program,
            coords=example_coords,
            priority=5,
        )  # Priority must be between 0 and 3


def test_target_invalid_priority_type(example_program, example_coords):
    with pytest.raises(TypeError):
        Target(
            name="Example Target",
            program=example_program,
            coords=example_coords,
            priority="High",
        )  # Priority must be an integer


def test_target_invalid_comment_type(example_program, example_coords):
    with pytest.raises(TypeError):
        Target(
            name="Example Target",
            program=example_program,
            coords=example_coords,
            comment=123,
        )  # Comment must be a string


def test_target_add_merit(example_target, example_merit):
    example_target.add_merit(example_merit)
    assert example_merit in example_target.veto_merits


def test_target_add_merit_invalid_type(example_target):
    with pytest.raises(TypeError):
        example_target.add_merit("InvalidMerit")  # Merit must be of type Merit


def test_target_add_merits(example_target, example_merit):
    merits = [example_merit]
    example_target.add_merits(merits)
    assert example_merit in example_target.veto_merits


def test_target_add_merits_invalid_list_type(example_target):
    with pytest.raises(TypeError):
        example_target.add_merits("InvalidMerits")  # Merits must be a list


def test_target_add_merits_invalid_item_type(example_target):
    with pytest.raises(TypeError):
        example_target.add_merits(
            ["InvalidMerit"]
        )  # All items in merits list must be of type Merit


def test_target_str_method(example_target):
    assert str(example_target).startswith("Target(Name: Example Target")


def test_target_repr_method(example_target):
    assert repr(example_target).startswith("Target(Name: Example Target")


def test_target_equality(example_target, example_program, example_coords):
    target1 = Target(
        name="Example Target",
        program=example_program,
        coords=example_coords,
        priority=1,
    )
    target2 = Target(
        name="Example Target",
        program=example_program,
        coords=example_coords,
        priority=2,
    )
    target3 = Target(
        name="Different Target",
        program=example_program,
        coords=example_coords,
        priority=1,
    )
    assert target1 == target2  # Same name, should be equal
    assert target1 != target3  # Different name, should not be equal


def test_target_equality_different_type(example_target):
    assert example_target != "Not a Target"  # Should not be equal to a different type



import pytest

from scopes.scheduler_components import Merit


@pytest.fixture
def example_merit_func():
    def custom_merit_function(observation, example_param, verbose=False):
        """Example merit function."""
        if verbose:
            print("Custom merit function called.")
        return example_param

    return custom_merit_function


def test_merit_initialization(example_merit_func):
    merit = Merit(
        name="Example Merit",
        func=example_merit_func,
        merit_type="efficiency",
        parameters={"example_param": 0.5},
    )
    assert merit.name == "Example Merit"
    assert merit.func == example_merit_func
    assert merit.description == "Example merit function."
    assert merit.merit_type == "efficiency"
    assert merit.parameters == {"example_param": 0.5}


def test_merit_invalid_merit_type(example_merit_func):
    with pytest.raises(ValueError):
        Merit(
            name="Invalid Merit",
            func=example_merit_func,
            merit_type="invalid_type",
            parameters={"example_param": 0.5},
        )


def test_merit_invalid_first_parameter_in_func():
    def invalid_merit_function(invalid_param, example_param):
        return example_param

    with pytest.raises(KeyError):
        Merit(
            name="Invalid Merit",
            func=invalid_merit_function,
            merit_type="efficiency",
            parameters={"example_param": 0.5},
        )


def test_merit_missing_required_parameters(example_merit_func):
    with pytest.raises(ValueError):
        Merit(
            name="Missing Params Merit",
            func=example_merit_func,
            merit_type="efficiency",
        )


def test_merit_extra_parameters(example_merit_func):
    with pytest.raises(KeyError):
        Merit(
            name="Extra Params Merit",
            func=example_merit_func,
            merit_type="efficiency",
            parameters={"example_param": 0.5, "extra_param": 1.0},
        )


def test_merit_evaluate_method(example_merit_func, example_observation):
    merit = Merit(
        name="Example Merit",
        func=example_merit_func,
        merit_type="efficiency",
        parameters={"example_param": 0.5},
    )
    result = merit.evaluate(example_observation)
    assert result == 0.5


def test_merit_evaluate_with_runtime_parameters(
    example_merit_func, example_observation
):
    merit = Merit(
        name="Example Merit",
        func=example_merit_func,
        merit_type="efficiency",
        parameters={"example_param": 0.5},
    )
    result = merit.evaluate(example_observation, verbose=True)
    assert result == 0.5


def test_merit_str_method(example_merit_func):
    merit = Merit(
        name="Example Merit",
        func=example_merit_func,
        merit_type="efficiency",
        parameters={"example_param": 0.5},
    )
    expected_str = "Merit(Example Merit, efficiency, {'example_param': 0.5})"
    assert str(merit) == expected_str


def test_merit_repr_method(example_merit_func):
    merit = Merit(
        name="Example Merit",
        func=example_merit_func,
        merit_type="efficiency",
        parameters={"example_param": 0.5},
    )
    expected_repr = "Merit(Example Merit, efficiency, {'example_param': 0.5})"
    assert repr(merit) == expected_repr



import pytest

from scopes.scheduler_components import Overheads


def test_overheads_initialization():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5, cable_wrap_angle=180.0)
    assert overheads.slew_rate_az == 1.0
    assert overheads.slew_rate_alt == 0.5
    assert overheads.cable_wrap_angle == 180.0
    assert overheads.overheads == []


def test_overheads_invalid_slew_rate():
    with pytest.raises(ValueError):
        Overheads(slew_rate_az=-1.0, slew_rate_alt=0.5)  # Invalid slew rate


def test_overheads_invalid_cable_wrap_angle():
    with pytest.raises(ValueError):
        Overheads(
            slew_rate_az=1.0, slew_rate_alt=0.5, cable_wrap_angle=400.0
        )  # Invalid angle


def test_overheads_validate_function_params():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5)

    def valid_func(observation1, observation2):
        pass

    def invalid_func(obs1, obs2):
        pass

    assert overheads._validate_function_params(valid_func)
    assert not overheads._validate_function_params(invalid_func)


def test_overheads_is_angle_between():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5)
    assert overheads._is_angle_between(10, 30, 20)
    assert overheads._is_angle_between(350, 10, 0)
    assert not overheads._is_angle_between(30, 10, 40)


def test_overheads_add_overhead():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5)

    def valid_func(observation1, observation2):
        return 100

    overheads.add_overhead(valid_func)
    assert len(overheads.overheads) == 1


def test_overheads_add_overhead_invalid_function():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5)

    def invalid_func(obs1, obs2):
        return 100

    with pytest.raises(ValueError):
        overheads.add_overhead(invalid_func)


def test_overheads_calculate_slew_time(
    example_full_observation1, example_full_observation2, example_overheads: Overheads
):
    slew_time = example_overheads.calculate_slew_time(
        example_full_observation1, example_full_observation2
    )
    assert slew_time > 0


def test_overheads_calculate_transition(
    example_full_observation1, example_full_observation2, example_overheads: Overheads
):
    def custom_overhead(observation1, observation2):
        return 100  # Mocking overhead time in seconds

    example_overheads.add_overhead(custom_overhead)
    total_overhead = example_overheads.calculate_transition(
        example_full_observation1, example_full_observation2
    )
    assert total_overhead > 100 / 86400  # Overhead time in days


def test_overheads_str_method():
    overheads = Overheads(slew_rate_az=1.0, slew_rate_alt=0.5, cable_wrap_angle=180.0)
    expected_str = "Overheads(\n\tslew_rate_az=1.0, \n\tslew_rate_alt=0.5, \n\tcable_wrap_limit=180.0)"
    assert str(overheads) == expected_str



from copy import deepcopy
from dataclasses import dataclass, field
from queue import PriorityQueue
from typing import Any, List, Tuple, Union

import numpy as np
from tqdm.auto import tqdm

from .merits import airmass_efficiency, end_time
from .scheduler_components import Merit, Night, Observation, Overheads, Plan


## ----- BASE SCHEDULER CLASS ----- ##
class Scheduler:
    """
    Base class for all schedulers.

    This class doesn't actually do any scheduling, but it sets up the necessary initializations
    and provides some helper functions that are common to all schedulers.
    """

    def __init__(
        self,
        night: Night,
        obs_list: List[Observation],
        overheads: Overheads,
        plan_start_time: float = None,
        plan_end_time: float = None,
    ) -> None:
        """
        Initializes the Scheduler class. Checks the validity of the plan start time and sets the
        start time of all observations. Also runs the skypath function for each observation.

        Parameters
        ----------
        night : Night
            The night object that defines the observable time.
        obs_list : List[Observation]
            The list of observations to schedule.
        overheads : Overheads
            The overheads object that defines the transition times between observations.
        plan_start_time : float, optional
            The start time of the plan in Julian Date. If None, it is set to the start of the
            observable night (as defined in the night object).
        plan_end_time : float, optional
            The end time of the plan in Julian Date. If None, it is set to the end of the
            observable night (as defined in the night object).
        """
        print("Preparing observations for scheduling...")
        # Set plan_start_time to the start of the observable night if it's None, otherwise use the given value
        self.plan_start_time = (
            night.obs_within_limits[0] if plan_start_time is None else plan_start_time
        )

        # Check that the selected plan start time is within the observable night limits
        if not (
            night.obs_within_limits[0]
            <= self.plan_start_time
            < night.obs_within_limits[1]
        ):
            raise ValueError(
                f"plan_start_time ({self.plan_start_time}) is outside the observable night limits "
                f"({night.obs_within_limits[0]} to {night.obs_within_limits[1]})."
            )
        # Set plan_end_time to the end of the observable night if it's None, otherwise use the given value
        custom_end_time = plan_end_time is not None
        self.plan_end_time = (
            night.obs_within_limits[1] if plan_end_time is None else plan_end_time
        )

        # Check that the selected plan end time is within the observable night limits
        if not (
            night.obs_within_limits[0]
            <= self.plan_end_time
            <= night.obs_within_limits[1]
        ):
            raise ValueError(
                f"plan_end_time ({self.plan_end_time}) is outside the observable night limits "
                f"({night.obs_within_limits[0]} to {night.obs_within_limits[1]})."
            )

        # Check that obs_list is a list of Observation objects
        if not isinstance(obs_list, list):
            raise TypeError("obs_list has to be a list of Observation objects.")
        # Check that obs_list is not empty and contains more than one observation
        if len(obs_list) <= 1:
            raise ValueError("obs_list has to contain more than one observation.")
        # Check that obs_list contains only Observation objects
        if not all(isinstance(obs, Observation) for obs in obs_list):
            raise TypeError("obs_list has to contain only Observation objects.")
        if not isinstance(night, Night):
            raise TypeError("night has to be a Night object.")
        if not isinstance(overheads, Overheads):
            raise TypeError("overheads has to be an Overheads object.")

        self.night = night
        self.obs_list = obs_list
        self.overheads = overheads
        # Set the start of all obs
        for obs in tqdm(self.obs_list, desc="Setting up observations"):
            if custom_end_time:
                obs.target.add_merit(
                    Merit("EndTime", end_time, "veto", {"time": self.plan_end_time})
                )
            # Set the night and start time
            obs.set_night(self.night)
            obs.set_start_time(self.plan_start_time)
            # Run skypath to calculate the path of the object during the night
            obs.skypath()
            obs.update_alt_airmass()

        # Calculate the extended time range for the culmination merit
        self.night.calculate_culmination_window(self.obs_list)

    def _check_max_plan_length(self, max_plan_length: Union[int, None]):
        """
        Checks the validity of the maximum plan length.

        Parameters
        ----------
        max_plan_length : int or None
            The maximum plan length. If None, it is set to the number of observations.

        Raises
        ------
        ValueError:
            If max_plan_length is greater than the number of observations or if it is not a
            positive integer or None.

        Returns
        -------
        max_plan_length : int
            The validated maximum plan length.
        """
        if max_plan_length is None:
            max_plan_length = len(self.obs_list)
        elif max_plan_length > len(self.obs_list):
            raise ValueError(
                "max_plan_length should be less than or equal to the number of observations."
            )
        if max_plan_length and max_plan_length <= 0:
            raise ValueError("max_plan_length should be a positive integer or None.")
        return max_plan_length

    def _obslist_deepcopy(self, obslist: List[Observation]):
        """
        An implementation of deepcopying a list of observations by creating new emtpy observations
        and assigning the attributes of the original observations to the new ones.

        This is a workaround for the fact that deepcopy is very slow for these types of objects.
        This alone can speed up the scheduling algorithms by a factor of 10.

        The only caveat is that this is not actually a deepcopy as the Target objects inside the
        Observation objects are not copied. This means is you make a modification on the Targets
        within these observation then this will affect the original observations as well.

        Parameters
        ----------
        obslist : List[Observation]
            The list of observations to deepcopy
        """
        new_obslist = []
        for obs in obslist:
            new_obs = Observation.__new__(Observation)
            new_obs.__dict__ = obs.__dict__.copy()
            new_obslist.append(new_obs)
        return new_obslist

    def _plan_deepcopy(self, plan: Plan):
        """
        An implementation of deepcopying a plan by creating new emtpy observations and assigning
        the attributes of the original observations to the new ones.

        This is a workaround for the fact that deepcopy is very slow for these types of objects.
        This alone can speed up the scheduling algorithms by a factor of 10.

        Parameters
        ----------
        plan : Plan
            The plan to deepcopy
        """
        # Create a new empty Plan
        new_plan = Plan()
        # Deepcopy the observations
        obs_copy = self._obslist_deepcopy(plan.observations)
        # Add the observations to the new plan
        for obs in obs_copy:
            new_plan.add_observation(obs)
        # Evaluate plan
        new_plan.evaluate_plan()
        return new_plan

    def update_start_times(
        self, observations: List[Observation], new_start_time: float
    ):
        """
        Update the start time of all observations in the list based on defined start time.

        Parameters
        ----------
        observations : List[Observation]
            The list of observations to update
        new_start_time : float
            The new start time to set for all observations
        """
        for obs in observations:
            obs.set_start_time(new_start_time)
            obs.update_alt_airmass()

    def update_start_from_prev(
        self, observations: List[Observation], previous_obs: Observation
    ):
        """
        Update the start time of all observations in the list based on the previous observation.

        Parameters
        ----------
        observations : List[Observation]
            The list of observations to update
        previous_obs : Observation
            The previous observation
        """
        for obs in observations:
            self.transition(previous_obs, obs)

    def transition(self, obs1: Observation, obs2: Observation):
        """
        Use the overheads class to calculate the transition time from obs1 to obs2. Then update
        the start time of obs2 and recalculate the score of obs2.

        Parameters
        ----------
        obs1 : Observation
            The first observation
        obs2 : Observation
            The second observation
        """
        total_overhead = self.overheads.calculate_transition(obs1, obs2)
        # Update the start time of obs2
        obs2.set_start_time(obs1.end_time + total_overhead)
        # Update the time array becaue the start time changed
        obs2.update_alt_airmass()
        # Calculate new rank score based on new start time
        obs2.feasible()
        obs2.evaluate_score()

    def move_observation(self, plan: Plan, index: int, new_index: int) -> Plan:
        """
        Move an observation to a new index in the plan.

        V2: Implement it more efficiently. Not as a series of swaps but directly placing the
        observation at the new location and moving all the observations in the middle as a block.

        Parameters
        ----------
        plan : Plan
            The plan to move the observation in
        index : int
            The index of the observation to move
        new_index : int
            The new index to move the observation to

        Returns
        -------
        Plan
            The plan with the moved observation
        """
        # Ensure indices are within the range of the plan's observations
        if index < 0 or index >= len(plan.observations):
            raise IndexError(f"index ({index}) is out of bounds.")
        if new_index < 0 or new_index >= len(plan.observations):
            raise IndexError(f"new_index {new_index} is out of bounds.")

        # If indices are the same there is no need to swap, return the original plan
        if index == new_index:
            return plan

        # Create a deep copy of the plan
        moved_plan = self._plan_deepcopy(plan)
        obss = moved_plan.observations

        # Move the observation to the new index
        moved_observation = obss.pop(index)
        obss.insert(new_index, moved_observation)

        left_idx = min(index, new_index)
        start_idx = left_idx - 1 if left_idx > 0 else 0
        # Update start time to end time of previous
        if start_idx == 0:
            # Beginning of plan, set start time to chosen start of plan
            obss[0].update_start_and_score(self.plan_start_time)
            obss[1].update_start_and_score(obss[0].end_time)
        else:
            obss[left_idx].update_start_and_score(obss[start_idx].end_time)
        # Check that observations didn't become unfeasible
        if obss[left_idx].score == 0.0:
            # If any of the scores are 0, the plan is infeasible, return the original plan
            return plan

        # Calculate new overheads for the transitions for following observations
        for idx in range(start_idx, len(obss) - 1):
            # Calculate new overhead
            # new_overhead = self.overheads.calculate_transition(obss[idx], obss[idx + 1])
            # # Update start time of next observation
            # obss[idx + 1].update_start_and_score(obss[idx].end_time + new_overhead)
            self.transition(obss[idx], obss[idx + 1])
            if obss[idx + 1].score == 0.0:
                # Observation becomes infeasible, abort and return the original plan
                return plan

        # Evaluate plan before returning
        moved_plan.evaluate_plan()

        return moved_plan

    def swap_observations(self, plan: Plan, index1: int, index2: int) -> Plan:
        """
        Swaps any two observations in the plan adjusts the start times and overheads for this
        new schedule.

        Parameters
        ----------
        plan : Plan
            The plan to swap observations in
        index1 : int
            The index of the first observation to swap
        index2 : int
            The index of the second observation to swap

        Returns
        -------
        Plan
            The plan with the swapped observations
        """
        # Ensure indices are within the range of the plan's observations
        if (
            index1 < 0
            or index1 >= len(plan.observations)
            or index2 < 0
            or index2 >= len(plan.observations)
        ):
            raise IndexError("Swap indices are out of bounds.")

        if index1 == index2:
            # No need to swap, return the original plan
            return plan

        # Create copy of the plan to not modify the original one
        plan_copy = self._plan_deepcopy(plan)

        # Swap the observations
        obss = plan_copy.observations
        obss[index1], obss[index2] = obss[index2], obss[index1]
        # Swap their start times and update the alt airmasses
        # Save the start time of obs1 before it gets changed
        obs_idx1_start = obss[index1].start_time
        obss[index1].update_start_and_score(obss[index2].start_time)
        obss[index2].update_start_and_score(obs_idx1_start)
        if not (obss[index2].score or obss[index1].score):
            # If any of the scores are 0, the plan is infeasible, return the original plan
            return plan

        # Calculate new overheads for the transitions between the previous and next observations
        # Calculate new transition for all obs from index1-1 until the end.
        left_idx = min(index1, index2)
        start_idx = left_idx - 1 if left_idx > 0 else 0
        for idx in range(start_idx, len(obss) - 1):
            # Calculate new overhead
            new_overhead = self.overheads.calculate_transition(obss[idx], obss[idx + 1])
            # Update start time of next observation
            obss[idx + 1].update_start_and_score(obss[idx].end_time + new_overhead)
            if obss[idx + 1].score == 0.0:
                # Observation becomes infeasible, abort and return the original plan
                return plan

        # Evaluate plan
        plan_copy.evaluate_plan()

        return plan_copy

    def _get_best(self, plans, metric, comparison_plan):
        """
        Get the best plan from a list of plans based on a given metric and comparison plan.

        Parameters
        ----------
        plan : Plan
            The plan to check
        metric : str
            The metric to use for comparison
        comparison_plan : Plan
            The plan to compare the other plans to

        Returns
        -------
        bool
            True if the plan is better, False otherwise
        """
        if metric == "overheads":
            overheads = [plan.overhead_time for plan in plans]
            best_plan = plans[np.argmin(overheads)]
            if best_plan.overhead_time < comparison_plan.overhead_time:
                return best_plan
        elif metric == "score":
            scores = [plan.score for plan in plans]
            best_plan = plans[np.argmax(scores)]
            if best_plan.score > comparison_plan.score:
                return best_plan
        else:
            raise ValueError("Invalid metric. Use 'overheads' or 'score'.")

        return comparison_plan

    def _get_metric(self, plan, metric):
        """
        Get the metric of a plan based on the given metric.

        Parameters
        ----------
        plan : Plan
            The plan to get the metric from
        metric : str
            The metric to get

        Returns
        -------
        float
            The value of the metric
        """
        if metric == "overheads":
            return plan.overhead_time
        elif metric == "score":
            return plan.score
        else:
            raise ValueError("Invalid metric. Use 'overheads' or 'score'.")

    def lsh_optimization(
        self,
        plan: Plan,
        method: str,
        metric: str,
        iterations: int = None,
        verbose: bool = False,
    ) -> Plan:
        """
        Optimize the plan using the LSH (Local Search Heuristic) algorithm. The algorithm works by
        going through all possible swaps or moves of observations and checking if the chosen metric
        is improved and takes the best one. If it is, the swap or move is accepted, otherwise it is
        rejected. The process is repeated for a number of iterations or until the metric doesn't
        improve. The algorithm can be used to optimize the plan based on the overhead time or the
        score of the plan. The method can be either "swap" or "move". Swap swaps two observations
        in the plan, while move moves an observation to a new spot in the plan.

        Parameters
        ----------
        plan : Plan
            The plan to optimize
        method : str
            The method to use for the optimization. Can be either "swap" or "move".
        metric : str
            The metric to use for the optimization. Can be either "overheads" or "score".
        iterations : int, optional
            The number of iterations to run the optimization for. If None, it runs until the score
            doesn't improve, by default None
        verbose : bool, optional
            If True, print the progress of the optimization, by default False

        Returns
        -------
        Plan
            The optimized plan
        """
        # Check that the method is either "swap" or "move" and assign the corresponding function
        if method not in ["swap", "move"]:
            raise ValueError("Method must be either 'swap' or 'move'.")
        if method == "move":
            operation_func = self.move_observation
        elif method == "swap":
            operation_func = self.swap_observations

        # Check that the metric is either "overheads" or "score"
        if metric not in ["overheads", "score"]:
            raise ValueError("Metric must be either 'overheads' or 'score'.")

        # Initialize the best plan to the original plan
        best_plan = self._plan_deepcopy(plan)
        if metric == "score":
            # Remove Culmination efficiency merit, and add the airmass efficiency merit
            for obs in best_plan.observations:
                # Check if the culmination_efficiency merit is present
                for merit in obs.target.efficiency_merits:
                    if "culmination" in merit.func.__name__.lower():
                        # Remove the culmination efficiency merit
                        obs.target.efficiency_merits.remove(merit)
                # Add the airmass efficiency merit
                obs.target.add_merit(
                    Merit(
                        "AirmassEfficiency",
                        airmass_efficiency,
                        merit_type="efficiency",
                    )
                )
                # Update scores with the new merit
                obs.feasible()
                obs.evaluate_score()
            # Evaluate the plan to recalculate score with new merits
            best_plan.evaluate_plan()

        if verbose:
            if metric == "overheads":
                print(f"Initial plan overhead time: {plan.overhead_time}")
            elif metric == "score":
                # Save initial scores and airmasses because of the change of merits
                intial_score = best_plan.score
                intial_avg_amass = best_plan.avg_airmass
                print(f"Initial plan score: {intial_score:.4f}")
                print(f"Initial avg airmass: {intial_avg_amass:.4f}")

        improved = True
        iteration_count = 0

        # TODO - Implement a tracking system to keep track of plans that were already rejected
        # This is to avoid recalculating the move operation of plans that were already done
        # rejected_states = {}

        # Make the forward and backward passes until the score doesn't improve
        while improved and (iterations is None or iteration_count < iterations):
            improved = False
            current_plan = self._plan_deepcopy(best_plan)
            # The indices of the observations to visit (variable U in the algorithm)
            pos_to_visit = list(range(len(plan.observations)))
            np.random.shuffle(pos_to_visit)  # Shuffles in-place
            move_count = 0
            for pos in pos_to_visit:
                # Go through all posible swaps and keep track of the overhead times
                plans = []
                for i in range(len(plan.observations)):
                    if i == pos:
                        # Skip the unnecesary swap with itself
                        continue
                    # Make the operation on the observation at pos
                    new_plan = operation_func(best_plan, pos, i)

                    if new_plan == best_plan:
                        # If the new plan is the same as the original plan, continue
                        # This happens when the swap is not feasible
                        continue
                    # Save the overhead time
                    plans.append(new_plan)

                # Find the plan with the lowest overhead time
                # Check that overheads is not empty
                if len(plans) == 0:
                    # If overheads is empty, there was no feasible swap, continue
                    continue
                else:
                    tmp_best_plan = self._get_best(plans, metric, best_plan)
                    # TODO idea: only accept the change if the time gain is "worth it"
                    # Where worth it has to be defined where we can look at the general score and if
                    # the score drop is too high to justify the time gain then the switch is not
                    # accepted. Or something like that...

                    if tmp_best_plan != best_plan:
                        # If the new plan is different from the original plan, accept the change
                        best_plan = tmp_best_plan
                        move_count += 1
                        improved = True

            # Update the iteration count
            iteration_count += 1

            if verbose:
                if metric == "overheads":
                    delta_t = current_plan.overhead_time - best_plan.overhead_time
                    print(f"\n----- Iteration {iteration_count} -----")
                    print(f"Overhead time: {best_plan.overhead_time}")
                    print(f"Time saved with {move_count} changes: {delta_t}\n")
                elif metric == "score":
                    delta_score = current_plan.score - best_plan.score
                    print(f"\n----- Iteration {iteration_count} -----")
                    print(f"Score: {best_plan.score:.4f}")
                    print(
                        f"Score improvement with {move_count} changes: {delta_score:.4f}\n"
                    )

        if verbose:
            print("Optimization finished.")
            if metric == "overheads":
                print(f"Optimized plan overhead time: {best_plan.overhead_time}")
                print(
                    "Total time saved: ", plan.overhead_time - best_plan.overhead_time
                )
            elif metric == "score":
                score_delta = intial_score - best_plan.score
                avg_amass_delta = intial_avg_amass - best_plan.avg_airmass
                print(
                    f"Optimized plan score: {best_plan.score:.4f} ({score_delta:+.4f})"
                )
                print(
                    f"Optimized plan avg airmass: {best_plan.avg_airmass:.4f} ({avg_amass_delta:+.4f})"
                )

        return best_plan


## ----- SPECIFIC SCHEDULERS ----- ##
class generateQ(Scheduler):
    def __init__(self, *args, **kwargs) -> None:
        # Call the parent class init
        super().__init__(*args, **kwargs)

    # Basic forward scheduler, greedy search
    def forwardP(
        self,
        start_obs: Union[Observation, float],
        available_obs: List[Observation],
        lookahead_distance=None,
    ):
        """
        Basic scheduler that simply continues a Plan from the starting observation by
        sequentially choosing the highest scoring observation.
        """

        # Set the lookahead distance to the number of available observations if not specified
        # Or to finish the night if there are more available observations than time in the night
        if (lookahead_distance is not None) and (
            len(available_obs) < lookahead_distance
        ):
            raise ValueError(
                f"Number of available observations ({len(available_obs)}) "
                f"must be more than or equal to lookahead distance ({lookahead_distance})"
            )
        elif lookahead_distance is None:
            # Equivalent to planning until the end of night or until no observations remain
            lookahead_distance = len(available_obs)

        # Initialize the Plan object according to if the starting codition is a Time or Observation
        observation_plan = Plan()
        if isinstance(start_obs, Observation):
            observation_plan.add_observation(start_obs)
            self.update_start_from_prev(available_obs, start_obs)
        elif isinstance(start_obs, float):
            self.update_start_times(available_obs, start_obs)

        else:
            raise TypeError(
                f"start_obs must be of type Observation or Time (as a float in jd), not {type(start_obs)}"
            )

        # Add candidate observation to plan K times

        for _ in range(lookahead_distance):
            # Initialize Q as an empty list to store ranked observations
            Q = []

            # Evaluate each available observation
            for o_prime in available_obs:
                if o_prime.feasible():
                    score = o_prime.evaluate_score()
                    # Insert into Q ensuring Q is sorted by score
                    Q.append((score, o_prime))

            # Sort Q by score
            Q.sort(reverse=True, key=lambda x: x[0])

            # Check exit conditions
            if not Q or len(observation_plan) >= lookahead_distance:
                break

            # Select the highest ranking observation
            if Q:
                # Select the highest ranking observation
                _, o_double_prime = Q[0]

                # Add the selected observation to the plan
                observation_plan.add_observation(o_double_prime)

                # Remove the selected observation from the available observations
                available_obs.remove(o_double_prime)

                # Update the start time of all remaining observations
                self.update_start_from_prev(available_obs, o_double_prime)

        # Evaluate the plan before returning
        observation_plan.evaluate_plan()

        return observation_plan

    def run(self, max_plan_length=None, K: int = 5):
        """
        The run function for the generateQ Scheduler. The way it works is by using
        the forwardP function to generate a plan from the starting time to the end of the night
        but at each step it does this for the top K observations, and it chooses the observation
        that has the highest plan score at the end of the night. That would become the first
        observation of the plan. After each step it performs a Local Search Heuristic optimization
        that optimizes the order of the observation to minimize overheads.
        Then for the second it repeats the same process. It takes the top K observations runs
        forwardP until the end of the night, and assesses the plan score (but of the entire night,
        including the observation that was added before). It then chooses the observation that has
        the highest plan score at the end of the night, and that becomes the second observation of
        the plan. It repeats this process until the plan is full or it reaches the maximum plan
        length.

        At the end a final LSH optimization is run to maximize the score of the overall plan. This
        usually means reordering the observation to achieve a higher average airmass for the
        observations.

        Parameters
        ----------
        max_plan_length : int, optional
            The maximum length of the plan, by default None meaning it will go until the end of
            the night
        K : int, optional
            The number of top observations to consider at each step, by default 5

        Returns
        -------
        Plan
            The final plan
        """
        print("Creating the Plan...")
        # Check max_plan_length
        max_plan_length = self._check_max_plan_length(max_plan_length)

        # Create an empty plan to store the final results
        final_plan = Plan()

        # Create a deep copy of the available observations
        remaining_obs: List[Observation] = self._obslist_deepcopy(self.obs_list)
        # Iterate until the plan is full or remaining_obs is empty
        while remaining_obs and (len(final_plan) < max_plan_length):
            # Score each observation to sort them and pick the top K
            obs_scores = []
            for o in remaining_obs:
                if o.feasible():
                    score = o.evaluate_score()
                    obs_scores.append((score, o))
            if not obs_scores:
                # No feasible observations remain
                break
            obs_scores.sort(reverse=True, key=lambda x: x[0])

            top_k_observations = [obs for _, obs in obs_scores[:K]]

            # Track the best observation and corresponding plan
            best_observation = None
            best_plan = None

            for obs in top_k_observations:
                # Create a deep copy of available_obs
                remaining_obs_copy = self._obslist_deepcopy(remaining_obs)
                # Remove current obs from the copy
                remaining_obs_copy.remove(obs)

                if K == 1:
                    # Ensure the lookahead distance is at least 1
                    lookahead_distance = max_plan_length - len(final_plan) - 1
                else:
                    # Set the lookahead distance to the minimum of 5 and the remaining observations
                    lookahead_distance = min(5, max_plan_length - len(final_plan) - 1)
                # Generate plan using forwardP with the modified list
                plan = self.forwardP(
                    obs,
                    remaining_obs_copy,
                    lookahead_distance=lookahead_distance,
                )

                # If the current plan is better than the best, update best_plan and best_observation
                if not best_plan or (plan.score > best_plan.score):
                    best_plan = plan
                    best_observation = obs

            if K == 1:
                # If K=1, the best plan is the one generated by forwardP
                final_plan = best_plan
                break
            else:
                # Add the best observation to the final plan
                final_plan.add_observation(best_observation)
                # Optimize the overheads of the plan
                final_plan = self.lsh_optimization(
                    final_plan, method="move", metric="overheads"
                )
                # Remove the best observation from the available observations list
                remaining_obs.remove(best_observation)
                # self.update_start_from_prev(remaining_obs, best_observation)
                self.update_start_from_prev(remaining_obs, final_plan.observations[-1])

        # Perform a final score optimization
        final_plan = self.lsh_optimization(final_plan, method="move", metric="score")

        # Evaluate the final plan
        final_plan.evaluate_plan()
        print("Done!")

        return final_plan


# Dynamic programming scheduler using recursion
class DPPlanner(Scheduler):
    def __init__(self, *args, **kwargs) -> None:
        # Call the parent class init
        super().__init__(*args, **kwargs)
        self.DP = {}
        self.total_counter = 0
        self.saved_state_counter = 0

    def reset_dp(self):
        self.DP = {}
        self.total_counter = 0
        self.saved_state_counter = 0

    def dp_recursion(
        self,
        remaining_observations: List[Observation],
        # current_time: Time,
        current_plan: Plan,
        max_plan_length: int,
        K: int = 5,
    ) -> Tuple[float, Plan]:
        """
        Dynamic programming recursive function to find the best plan from a given state.
        It uses beam search to consider only the top K observations at each step.
        """
        self.total_counter += 1
        # Sort remaining_observations by target names
        sorted_remaining_observations = sorted(
            remaining_observations, key=lambda obs: obs.target.name
        )

        # Create the state tuple using sorted remaining observations and current plan length
        state = (
            tuple(obs.unique_id for obs in sorted_remaining_observations),
            len(current_plan),
        )

        # Check if state has already been computed
        if state in self.DP:
            self.saved_state_counter += 1
            return self.DP[state]

        # Base case 1: No remaining observations, evaluate current plan
        # Base case 2: Plan has reached maximum length
        if len(remaining_observations) == 0 or len(current_plan) >= max_plan_length:
            score = current_plan.evaluate_plan()
            self.DP[state] = (score, current_plan)
            return score, current_plan

        # Initialize variables to hold the best score and corresponding plan
        best_score = float("-inf")
        best_plan = Plan()

        # Evaluate feasibility and score of each observation
        for obs in remaining_observations:
            obs.feasible()
            obs.evaluate_score()
        # Select the top K observations to consider
        top_k_observations = sorted(
            remaining_observations, key=lambda x: x.score, reverse=True
        )[:K]

        # Loop through the top K observations to consider adding each to the plan
        for obs in top_k_observations:
            # Create a deep copy of current_plan first
            new_plan = deepcopy(current_plan)

            # Create a copy of the remaining observations
            remaining_copy = self._obslist_deepcopy(remaining_observations)
            # Remove the observation
            remaining_copy.remove(obs)

            # Check if adding this observation is feasible
            # NOTE: I think this check can be omitted as its already done in top_k_observations
            if obs.score > 0.0:
                # Add observation to the new plan
                new_plan.add_observation(obs)

                # Update the current time based on the end time of the added observation
                self.update_start_from_prev(remaining_copy, obs)

                # Recursive call to find best plan from this point forward
                _, temp_plan = self.dp_recursion(
                    remaining_copy, new_plan, max_plan_length, K
                )

                # Evaluate this complete plan
                score = temp_plan.evaluate_plan()

                # Update best score and plan if this plan is better
                if score > best_score:
                    best_score = score
                    best_plan = temp_plan

        # Store the best score and plan for this state
        self.DP[state] = (best_score, best_plan)

        return best_score, best_plan

    def run(
        self,
        max_plan_length: int = None,
        K: int = 5,
    ) -> Plan:
        # Check max_plan_length
        max_plan_length = self._check_max_plan_length(max_plan_length)

        for obs in self.obs_list:
            obs.feasible()

        # Call the recursive function to find the best plan
        _, best_plan = self.dp_recursion(self.obs_list, Plan(), max_plan_length, K)

        return best_plan


# Beam search scheduler
class BeamSearchPlanner(Scheduler):
    def __init__(self, *args, **kwargs) -> None:
        # Call the parent class init
        super().__init__(*args, **kwargs)
        self.total_counter = 0
        self.depth = 0

    @dataclass(order=True)
    class PrioritizedItem:
        score: float
        plan: Any = field(compare=False)
        obs: Any = field(compare=False)

    def run(
        self,
        max_plan_length: int = None,
        K: int = 5,
    ) -> Plan:
        # Check max_plan_length
        max_plan_length = self._check_max_plan_length(max_plan_length)

        # Initialize two priority queues
        PQ_current: PriorityQueue = PriorityQueue()
        PQ_next: PriorityQueue = PriorityQueue()

        # Add initial state to the current priority queue
        PQ_current.put(self.PrioritizedItem(0, Plan(), self.obs_list))

        # Initialize best plan and best score to None and -inf
        best_plan: Plan = Plan()

        while not PQ_current.empty():
            self.total_counter += 1

            # Retrieve the highest-score plan from the current priority queue
            pq_current_item = PQ_current.get()
            # current_score = pq_current_item.score
            current_plan: Plan = pq_current_item.plan
            remaining_observations: List[Observation] = pq_current_item.obs

            # Check stopping criteria
            if len(current_plan) >= max_plan_length:
                break

            Q = []
            # Generate child plans by extending the current plan with feasible observations
            for obs in remaining_observations:
                if obs.feasible():
                    score = obs.evaluate_score()
                    Q.append((score, obs))
            if not Q:
                # Termination condition if no feasible observations remain
                break

            # Sort Q by score
            Q.sort(reverse=True, key=lambda x: x[0])
            for _, obs in Q:
                # Copy of the plan
                new_plan = Plan()
                new_plan.observations = current_plan.observations[:]
                new_plan.add_observation(obs)
                # Copy of remaining obs
                new_remaining = self._obslist_deepcopy(remaining_observations)
                new_remaining.remove(obs)

                self.update_start_from_prev(new_remaining, obs)
                new_score = new_plan.evaluate_plan()
                PQ_next.put(self.PrioritizedItem(-new_score, new_plan, new_remaining))

            # If PQ_current is empty, move top-K from PQ_next to PQ_current
            if PQ_current.empty():
                # print(f"Current best plan: {best_current_plan}")
                self.depth += 1
                # Put top-K plans in the PQ_current queue
                best_score = 0.0
                for _ in range(min(K, PQ_next.qsize())):
                    pq_current_item = PQ_next.get()
                    # Update the best plan if this one is better
                    if pq_current_item.score < -best_score:
                        best_plan = pq_current_item.plan
                        best_score = pq_current_item.score
                    PQ_current.put(pq_current_item)
                # Update score of best plan

                # Clear PQ_next for the next iteration
                PQ_next = PriorityQueue()

        return best_plan


# Genetic algorithm scheduler
# class GeneticAlgorithm:
#     def __init__(self) -> None:
#         self.total_counter: int = 0

#     def nsga2(
#         self,
#         available_observations: List[Observation],
#         max_plan_length: int,
#         population_size: int,
#         generations: int,
#     ):
#         """Non-dominated Sorting Genetic Algorithm II (NSGA-II) implementation"""
#         return None



import uuid

import pytest

from scopes.scheduler_components import Night, Observation


def test_observation_initialization(example_target):
    observation = Observation(target=example_target, exposure_time=3600)
    assert observation.target == example_target
    assert observation.exposure_time == 3600 / 86400  # exposure time in days
    assert observation.tel_alt_lower_lim == 10.0
    assert observation.tel_alt_upper_lim == 90.0
    assert observation.score == 0.0
    assert isinstance(observation.unique_id, uuid.UUID)


def test_observation_invalid_target_type():
    with pytest.raises(TypeError):
        Observation(target="InvalidTarget", exposure_time=3600)


def test_observation_invalid_exposure_time_value(example_target):
    with pytest.raises(ValueError):
        Observation(target=example_target, exposure_time=0)  # exposure_time must be > 0


def test_observation_warning_for_short_exposure_time(example_target):
    with pytest.warns(UserWarning):
        Observation(
            target=example_target, exposure_time=0.5
        )  # Warning for exposure time < 1 second


def test_observation_warning_for_long_exposure_time(example_target):
    with pytest.warns(UserWarning):
        Observation(
            target=example_target, exposure_time=32401
        )  # Warning for exposure time > 9 hours


def test_observation_set_night(example_observation: Observation, example_night):
    example_observation.set_night(example_night)
    assert example_observation.night == example_night


def test_observation_set_start_time(example_observation: Observation):
    start_time = 2456789.5  # Example Julian Date
    example_observation.set_start_time(start_time)
    assert example_observation.start_time == start_time
    assert (
        example_observation.end_time == start_time + example_observation.exposure_time
    )


def test_observation_skypath(example_observation: Observation, example_night: Night):
    example_observation.set_night(example_night)
    example_observation.set_start_time(example_night.night_time_range[0].jd)
    example_observation.skypath()
    assert hasattr(example_observation, "night_altitudes")
    assert hasattr(example_observation, "night_azimuths")
    assert hasattr(example_observation, "night_airmasses")
    assert hasattr(example_observation, "min_altitude")
    assert hasattr(example_observation, "max_altitude")
    assert hasattr(example_observation, "culmination_time")
    assert hasattr(example_observation, "rise_time")
    assert hasattr(example_observation, "set_time")


def test_observation_fairness_no_merits(example_observation: Observation):
    example_observation.target.fairness_merits.clear()  # Ensure no fairness merits
    assert example_observation.fairness() == 1.0  # No merits, fairness should be 1.0


def test_observation_fairness_with_merits(
    example_observation_with_fairness: Observation,
):
    assert (
        example_observation_with_fairness.fairness() == 1.2
    )  # Should match the merit function's return value


def test_observation_update_alt_airmass(
    example_observation: Observation, example_night: Night
):
    example_observation.set_night(example_night)
    example_observation.set_start_time(example_night.night_time_range[0].jd)
    example_observation.skypath()  # Must be run to initialize skypath data
    example_observation.update_alt_airmass()
    assert len(example_observation.obs_time_range) > 0
    assert len(example_observation.obs_altitudes) > 0
    assert len(example_observation.obs_azimuths) > 0
    assert len(example_observation.obs_airmasses) > 0


def test_observation_feasible_no_merits(example_observation: Observation):
    example_observation.target.veto_merits.clear()  # Ensure no veto merits
    assert (
        example_observation.feasible() == 1.0
    )  # No veto merits, sensibility should be 1.0


def test_observation_feasible_with_veto_merit(
    example_observation_with_veto: Observation,
):
    assert (
        example_observation_with_veto.feasible() == 0.0
    )  # Should match the veto merit function's return value


def test_observation_evaluate_score_with_fairness(
    example_observation_with_fairness: Observation,
):
    example_observation_with_fairness.feasible()  # sensibility must be calculated first
    score = example_observation_with_fairness.evaluate_score()
    assert (
        score == 1.2
    )  # Since fairness = 1.2, sensibility = 1, efficiency = 1 (default)


def test_observation_evaluate_score_with_efficiency(
    example_observation_with_efficiency: Observation,
):
    example_observation_with_efficiency.set_start_time(2456789.5)
    example_observation_with_efficiency.feasible()  # sensibility must be calculated first
    score = example_observation_with_efficiency.evaluate_score()
    assert (
        score == 0.9
    )  # Since fairness = 1 (default), sensibility = 1, efficiency = 0.9


def test_observation_evaluate_score_with_veto(
    example_observation_with_veto: Observation,
):
    example_observation_with_veto.set_start_time(2456789.5)
    example_observation_with_veto.feasible()  # sensibility must be calculated first
    score = example_observation_with_veto.evaluate_score()
    assert score == 0.0  # Since veto merits should zero out the score


def test_observation_update_start_and_score(
    example_observation: Observation, example_night: Night
):
    start_time = 2460524.7  # Example Julian Date where the example target is visible
    example_observation.set_night(example_night)
    example_observation.set_start_time(start_time - 0.1)
    example_observation.skypath()
    example_observation.update_start_and_score(start_time)
    assert example_observation.start_time == start_time
    assert example_observation.score > 0.0  # Assume some score is calculated


def test_observation_str_method(example_observation: Observation):
    example_observation.set_start_time(2456789.5)
    expected_str = (
        f"Observation(Target: {example_observation.target.name},\n"
        f"            Start time: {example_observation.start_time},\n"
        f"            Exposure time: {example_observation.exposure_time},\n"
        f"            Score: {example_observation.score})"
    )
    assert str(example_observation) == expected_str


def test_observation_repr_method(example_observation: Observation):
    example_observation.set_start_time(2456789.5)
    expected_repr = (
        f"Observation(Target: {example_observation.target.name},\n"
        f"            Start time: {example_observation.start_time},\n"
        f"            Exposure time: {example_observation.exposure_time},\n"
        f"            Score: {example_observation.score})"
    )
    assert repr(example_observation) == expected_repr


def test_observation_equality(example_target):
    observation1 = Observation(target=example_target, exposure_time=3600)
    observation2 = Observation(target=example_target, exposure_time=3600)
    assert observation1 != observation2  # Different unique IDs, should not be equal
    assert observation1 == observation1  # Same object, should be equal


def test_observation_equality_different_type(example_observation):
    assert (
        example_observation != "Not an Observation"
    )  # Should not be equal to a different type



import os
import tempfile
from unittest import mock

import pandas as pd
import pytest

from scopes.scheduler_components import Plan


def test_plan_initialization():
    plan = Plan()
    assert len(plan.observations) == 0
    assert plan.score == 0.0
    assert plan.evaluation == 0.0


def test_plan_add_observation(example_full_observation1):
    plan = Plan()
    plan.add_observation(example_full_observation1)
    assert len(plan.observations) == 1
    assert plan.observations[0] == example_full_observation1


def test_plan_calculate_overhead(example_plan: Plan):
    example_plan.calculate_overhead()
    # 3 observations, each 600 seconds
    assert example_plan.observation_time.total_seconds() == 3 * 600
    # Overhead time should be positive
    assert example_plan.overhead_time.total_seconds() > 0
    # Observation ratio should be positive
    assert example_plan.observation_ratio > 0
    # Overhead ratio should be positive
    assert example_plan.overhead_ratio > 0


def test_plan_calculate_avg_airmass(example_plan: Plan):
    example_plan.calculate_avg_airmass()
    assert example_plan.avg_airmass > 0  # Average airmass should be positive


def test_plan_evaluate_plan(example_plan: Plan):
    evaluation = example_plan.evaluate_plan(w_score=0.4, w_overhead=0.6)
    # Evaluation should be positive
    assert evaluation > 0
    # Plan score should be calculated
    assert example_plan.score > 0
    # Evaluation should match the returned value
    assert example_plan.evaluation == evaluation


def test_plan_evaluate_plan_invalid_weights(example_plan):
    # Weights do not sum to 1
    with pytest.raises(ValueError):
        example_plan.evaluate_plan(w_score=0.5, w_overhead=0.6)


def test_plan_print_stats(example_plan: Plan):
    # This method just prints; no assertion needed
    example_plan.print_stats()


@pytest.mark.parametrize("display", [True, False])
def test_plan_plot(example_plan: Plan, display):
    example_plan.evaluate_plan()
    with mock.patch("matplotlib.pyplot.show") as mock_show:
        with mock.patch("matplotlib.pyplot.close") as mock_close:
            example_plan.plot(display=display)
            if display:
                mock_show.assert_called_once()
            else:
                mock_close.assert_called_once()


@pytest.mark.parametrize("path", [None, "test_plot.png"])
def test_plan_plot_save(example_plan: Plan, path):
    example_plan.evaluate_plan()
    with mock.patch("matplotlib.pyplot.savefig") as mock_savefig:
        example_plan.plot(display=False, path=path)
        if path is not None:
            mock_savefig.assert_called_once_with(path, dpi=300)


def test_plan_plot_interactive(example_plan: Plan):
    with mock.patch("plotly.graph_objs.Figure.show") as mock_show:
        example_plan.plot_interactive()
        mock_show.assert_called_once()


def test_plan_plot_polar(example_plan: Plan):
    with mock.patch("matplotlib.pyplot.show") as mock_show:
        example_plan.plot_polar(display=True)
        mock_show.assert_called_once()


def test_plan_plot_polar_save(example_plan: Plan):
    with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as tmpfile:
        path = tmpfile.name
        example_plan.plot_polar(display=False, path=path)

    assert os.path.exists(path) and os.path.getsize(path) > 0
    os.remove(path)


def test_plan_plot_altaz(example_plan: Plan):
    with mock.patch("matplotlib.pyplot.show") as mock_show:
        example_plan.plot_altaz(display=True)
        mock_show.assert_called_once()


def test_plan_plot_altaz_save(example_plan: Plan):
    with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as tmpfile:
        path = tmpfile.name
        example_plan.plot_altaz(display=False, path=path)

    assert os.path.exists(path) and os.path.getsize(path) > 0
    os.remove(path)


def test_plan_to_csv(example_plan: Plan):
    # Create a temporary file to save the CSV output
    with tempfile.NamedTemporaryFile(suffix=".csv", delete=False) as tmpfile:
        path = tmpfile.name
        example_plan.to_csv(path)

    # Check that the file is created and is non-empty
    assert os.path.exists(path) and os.path.getsize(path) > 0

    # Clean up the temporary file
    os.remove(path)


def test_plan_to_csv_with_custom_options(example_plan: Plan):
    # Create a temporary file to save the CSV output
    with tempfile.NamedTemporaryFile(suffix=".csv", delete=False) as tmpfile:
        path = tmpfile.name
        example_plan.to_csv(path, sep=";", index=False)  # Custom options for to_csv

    # Check that the file is created and is non-empty
    assert os.path.exists(path) and os.path.getsize(path) > 0

    # Read the CSV content with custom options and check
    df_from_csv = pd.read_csv(path, sep=";")
    assert not df_from_csv.empty

    # Clean up the temporary file
    os.remove(path)


def test_plan_len_method(example_plan):
    assert len(example_plan) == 3  # Length should match the number of observations


def test_plan_repr_method(example_plan):
    assert repr(example_plan) == f"<Plan> with {len(example_plan)} observations"


def test_plan_str_method(example_plan: Plan):
    df_str = example_plan.to_df().to_string()
    assert str(example_plan) == df_str



import re

from setuptools import find_packages, setup

# Read the contents of your README file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

# Read version directly from the package __init__.py file
with open("scopes/__init__.py", "r") as f:
    version = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)

setup(
    name="scopes-astro",
    version=version,
    author="Nicolas Unger",
    author_email="nicolas.unger@unige.ch",
    description="System for Coordinating Observational Planning and Efficient Scheduling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    package_dir={"scopes": "scopes"},
    install_requires=[
        "numpy",
        "pandas",
        "matplotlib",
        "astropy",
        "astroplan",
        "tqdm",
        "pytz",
        "timezonefinder",
    ],
    extras_require={"dev": ["pytest", "sphinx"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    license="GNU General Public License v3.0",
    keywords="astronomy scheduling observation planning",
    project_urls={
        "GitHub": "https://github.com/nicochunger/SCOPES",
        "ReadTheDocs": "https://scopes-docs.readthedocs.io/en/latest/",
        "Documentation": "https://github.com/nicochunger/SCOPES/blob/main/SCOPES_documentation.pdf",
        "Changelog": "https://github.com/nicochunger/SCOPES/blob/main/CHANGELOG.txt",
    },
)
